
#include	<windows.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>

#include "vtparser.h"
#include "vtprnctl.h"
#include "vtdcs.h"
#include "vt.h"
#include "emulator.h"

PS	vt_ps;

void	emu_output_str(LPSTR data, int len)
{
    ParseAnsiData(vt_ps, NULL, data, len);
}

void	emu_output(int ch)
{
    unsigned char	data = ch;

    ParseAnsiData(vt_ps, NULL, &data, 1);
}

void	debug(char *fmt, ...)
{
   char		buf[256];
   char		*p;
   va_list 	argptr;

   va_start(argptr, fmt);
   vsprintf(buf, fmt, argptr);
   va_end(argptr);

   p = buf;

   while (*p)
	tty_output(*p++);
}

static int	in_graphics = FALSE;

static void	set_attribute(int a)
{
    int		curr_attr = tty_get_attr();
    int		requested = 0;

    if (a == ATTR_RESET)
	curr_attr = 0;

    if (a & ATTR_BOLD)
	requested |= ZATTR_BOLD;

    if (a & ATTR_BLINK)
	requested |= ZATTR_BLINK;

    if (a & ATTR_UNDER)
	requested |= ZATTR_UNDERLINE;

    if (a & ATTR_NEG)
	requested |= ZATTR_REVERSE;

    if (a & ATTR_NOT)
	curr_attr &= ~requested;
    else
	curr_attr |= requested;

    tty_set_attr(curr_attr);

    if (a == ATTR_RESET)
    {
	tty_background_colour(-1);
	tty_foreground_colour(-1);
    }
}

typedef struct	t_fkeys
{
    int		key;
    char	*name;
}
    t_fkeys;

t_fkeys         Vt220_keys[] =
{
    {	ZKEY_BACKSPACE,	"\x007f"	},	/* DEL for backspace */
    {	ZKEY_UP,	"\x001b[A"	},
    {	ZKEY_DOWN,	"\x001b[B"	},
    {	ZKEY_RIGHT,	"\x001b[C"	},
    {	ZKEY_LEFT,	"\x001b[D"	},

    {	ZKEY_FIND,	"\x001b[1~"	},
    {	ZKEY_INSERT,	"\x001b[2~"	},
    {	ZKEY_REMOVE,	"\x001b[3~"	},
    {	ZKEY_SELECT,	"\x001b[4~"	},
    {	ZKEY_PGUP,	"\x001b[5~"	},
    {	ZKEY_PGDOWN,	"\x001b[6~"	},

    {	ZKEY_PF1,	"\x001bOP"	},
    {	ZKEY_PF2,	"\x001bOQ"	},
    {	ZKEY_PF3,	"\x001bOR"	},
    {	ZKEY_PF4,	"\x001bOS"	},

    {	ZKEY_F1,	"\x001b[11~"	},
    {	ZKEY_F2,	"\x001b[12~"	},
    {	ZKEY_F3,	"\x001b[13~"	},
    {	ZKEY_F4,	"\x001b[14~"	},
    {	ZKEY_F5,	"\x001b[15~"	},
    {	ZKEY_F6,	"\x001b[17~"	},
    {	ZKEY_F7,	"\x001b[18~"	},
    {	ZKEY_F8,	"\x001b[19~"	},
    {	ZKEY_F9,	"\x001b[20~"	},
    {	ZKEY_F10,	"\x001b[21~"	},
    {	ZKEY_F11,	"\x001b[23~"	},
    {	ZKEY_F12,	"\x001b[24~"	},
    {	ZKEY_F13,	"\x001b[25~"	},
    {	ZKEY_F14,	"\x001b[26~"	},
    {	ZKEY_F15,	"\x001b[28~"	},
    {	ZKEY_F16,	"\x001b[29~"	},
    {	ZKEY_F17,	"\x001b[31~"	},
    {	ZKEY_F18,	"\x001b[32~"	},
    {	ZKEY_F19,	"\x001b[33~"	},
    {	ZKEY_F20,	"\x001b[34~"	},
    0
};

void	emu_init(void)
{
    t_fkeys	*q;

    vt_ps = AllocateParser();

    for (q = Vt220_keys; q->key != 0; q++)
	key_define(q->key, q->name);
}

void 	 ProcessAnsi( LPSESINFO pSesptr, PA_INT code, LPARGS args )
{
    int		i;

    if (code > 0 && code < 256)
    {
	if (in_graphics)
	{
	    switch (code)
	    {
	    case 'x':
		code = ZGR_V_LINE;
		break;

	    case 'q':
		code = ZGR_H_LINE;
		break;

	    case 'l':
		code = ZGR_TL_CORNER;
		break;

	    case 'k':
		code = ZGR_TR_CORNER;
		break;

	    case 'm':
		code = ZGR_BL_CORNER;
		break;

	    case 'j':
		code = ZGR_BR_CORNER;
		break;

	    case 'u':
		code = ZGR_TEE_LEFT;
		break;

	    case 't':
		code = ZGR_TEE_RIGHT;
		break;

	    case 'v':
		code = ZGR_TEE_TOP;
		break;

	    case 'w':
		code = ZGR_TEE_BOTTOM;
		break;

	    case 'n':
		code = ZGR_CROSS;
		break;

	    default:
		code = ' ';
	    }
	}
	tty_output(code);
	return;
    }

    switch (code)
    {
    case DO_VTERR:	/* Oops. Internal parser error. Args[0]: error code.
                            If you think it is an implementation error
                            (codes 0 - 255, 257 and possibly 256 and 258),
                            please report to dosdev@mit.edu

                            DEBUG mode is on if _DEBUG or _PA_DEBUG symbols are
                              defined.

                            Error codes:
                              0 - 255 -> a char that has "escaped" getNextChar
                                          (DEBUG mode only)
                              256     -> an entry in vtcsi.tbl requests too
                                          many parameters (DEBUG mode only)
                              257     -> large buffer is not freed
                                          (DEBUG mode only)
                              258     -> an invalid mode in DECRSCI
                                          (DEBUG mode only)
                              512     -> out of memory
                              513     -> overflow. not enough buffer */
	return;

    case DO_ENQ:	/* Enquiry */
	/* FIXME */
	return;

    case DO_BEL:
	MessageBeep(0);
	return;

    case DO_LS0:	/* Lock shift G0. G0 --> GL */
	in_graphics = FALSE;
	return;

    case DO_LS1:	/* Lock shift G1. G1 --> GL */
	in_graphics = TRUE;
	return;

    case DO_LS1R:	/* Lock shift G1 right. G1 --> GR */
	/* FIXME */
	break;

    case DO_LS2:	/* Lock shift G2. G2 --> GL */
	/* FIXME */
	break;

    case DO_LS2R:	/* Lock shift G2 right. G2 --> GR */
	/* FIXME */
	break;

    case DO_LS3:	/* Lock shift G3. G3 --> GL */
	/* FIXME */
	break;

    case DO_LS3R:	/* Lock shift G3 right. G3 --> GR */
	/* FIXME */
	break;

    case DO_SS2:	/* Single shift G2. G2 --> GL for next graphic char */
	/* FIXME */
	break;

    case DO_SS3:	/* Single shift G3. G3 --> GL for next graphic char */
	/* FIXME */
	break;

    case DO_SCS:	/* Select character set.
                          Args[0]: n, where Gn is G0-G4
                          Args[1]: either CS_96 or CS_94: char set length
                          Args[2]: final char (see CS_xx in vt_const.h for the
                             final chars for the standard hard char sets)
                          Args[3]: encoded intermediate chars.  Formula:
                             Args[3] = #_of_intermediates * 0x0100 +
                                       (intermediate_#_1 - ' ') * 0x0010 +
                                       (intermediate_#_2 - ' ')
                           NOTE: You do NOT need to decode this argument
                           since the parser is always using the same encoding
                           in all functions */
	/* FIXME  - Not yet implemented */
	return;

    case DO_XON:
    case DO_XOFF:
	/* FIXME */
	return;

    case DO_KAM:	/* Keyboard Lock.  SET: locked; RESET: Unlocked */
	/* FIXME */
	break;

    case DO_IRM:	/* Insert Mode.  SET: Insert mode; RESET: Replace */
	if (vt_ps->iArgs[0])
	    tty_set_insert_mode(TRUE);
	else
	    tty_set_insert_mode(FALSE);
	return;

    case DO_SRM:	/* Local echo.  SET: OFF; RESET: ON */
	/* FIXME */
	break;

    case DO_LNM:	/* LF/New Line mode.
                          SET: Received LF, FF, or VT cause New Line action,
                            and pressed [Enter] or [Return] sends CR LF;
                          RESET: Received LF, FF, or VT cause cursor moving
                            to the next line on the same column, and pressed
                            [Enter] or [Return] sends CR */
	/* FIXME  - Not yet implemented */
	return;

    case DO_DECCKM:	/* Cursor key mode: SET: Application Mode;
                                            RESET: Cursor Mode */
	/* FIXME  - Not yet implemented */
	return;

    case DO_DECCOLM:	/* Columns: SET: 132 Columns; RESET: 80 columns */
	/* FIXME  - Not yet implemented */
	return;

    case DO_DECSCLM:	/* Scrolling. SET: Smooth scrolling;
                                    RESET: Jump scrolling */
	/* FIXME  - Not yet implemented */
	return;

    case DO_DECSCNM:	/* Reverse Screen.  SET: Reverse Screen;
                                            RESET: Normal Screen */
	if (vt_ps->iArgs[0])
	    tty_set_attr(tty_get_attr() | ZATTR_REVERSE);
	else
	    tty_set_attr(tty_get_attr() & ~ZATTR_REVERSE);
	return;

    case DO_DECOM:	/* Origin.  SET: Origin Mode; RESET: Absolute Mode */
	/* FIXME */
	break;

    case DO_DECAWM:	/* Auto wrap mode.  SET: ON; RESET: OFF */
	/* FIXME  - Not yet implemented */
	return;

    case DO_DECARM:	/* Auto repeat mode.  SET: ON; RESET: OFF */
	/* FIXME  - Not yet implemented */
	return;

    case DO_DECPFF:	/* Print FF mode.  SET: ON; RESET: OFF */
	/* FIXME */
	break;

    case DO_DECPEX:	/* Print extent mode. SET: Full Screen
                                              RESET: Scrolling region */
	/* FIXME */
	break;

    case DO_DECTCEM:	/* Text cursor.  SET: ON; RESET: OFF */
	if (vt_ps->iArgs[0])
	    tty_cursor_on();
	else
	    tty_cursor_off();
	return;

    case DO_DECNKM:	/* Keypad Mode: SET: Application; RESET: Numeric */
	/* FIXME  - Not yet implemented */
	return;

    case DO_DECNRCM:	/* Char Set. SET: National; RESET: Multinational */
	/* FIXME */
	break;

    case DO_GRM:	/* Char Set. SET: Use Special Graphics character set;
                                     RESET: Use Standard ASCII character set */
	/* FIXME */
	break;

    case DO_CRM:	/* Control Codes. SET: Act upon;
                                          RESET: Debug display */
	/* FIXME */
	break;

    case DO_HEM:	/* Horizontal Editing.  SET: ON; RESET: OFF */
	/* FIXME */
	break;

    case DO_DECINLM:	/* Interlace.  SET: ON; RESET: OFF */
	/* FIXME */
	break;

    case DO_DECKBUM:	/* Typewriter mode. SET: Data process;
                                            RESET: typewriter */
	/* FIXME */
	break;

    case DO_DECVSSM:	/* Vertical split screen mode. SET: ON; RESET: OFF */
	/* FIXME */
	break;

    case DO_DECSASD:	/* Select active status display. SET: Status Line
                                                         RESET: Main Display */
	/* FIXME */
	break;

    case DO_DECBKM:	/* Backarrow mode.
                          SET: backarrow is backspace and sends BS;
                          RESET: backarrow is delete and sends DEL  */
	/* FIXME */
	break;

    case DO_DECKPM:	/* Key position mode.
                          SET: send key position reports;
                          RESET: send character codes */
	/* FIXME */
	break;

    case DO_DECARSM:	/* Auto-resize mode. SET: on; RESET: off */
	/* FIXME */
	break;


    case DO_CUU:	/* Cursor UP  Args[0]: lines */
	tty_cursor_up(vt_ps->iArgs[0]);
	return;

    case DO_CUD:	/* Cursor DOWN  Args[0]: lines */
	tty_cursor_down(vt_ps->iArgs[0]);
	return;

    case DO_CUF:	/* Cursor right  Args[0]: columns */
	tty_cursor_right(vt_ps->iArgs[0]);
	return;

    case DO_CUB:	/* Cursor left  Args[0]: columns */
	tty_cursor_left(vt_ps->iArgs[0]);
	return;

    case DO_CUP:	/* Cursor move  Args[0]: line#  Args[1]: column# */
	tty_move_cursor(vt_ps->iArgs[1], vt_ps->iArgs[0]);
	return;

    case DO_IND:	/* Index. Move cursor down. Possible scroll up */
	tty_cursor_down_scroll();
	return;

    case DO_RI:	/* Reverse index. Possible scroll down */
	tty_cursor_up_scroll();
	return;

    case DO_NEL:	/* New Line. Move cursor 1st position on next line.
                            Possible scroll up */
	tty_cursor_down_scroll();
	/* FIXME */
	break;

    case DO_HOME:	/* Move Cursor to home */
	tty_cursor_home();
	return;

    case DO_UP:	/* Move up (no scrolling) */
	tty_cursor_up(1);
	return;

    case DO_DOWN:	/* Move down (no scrolling) */
	tty_cursor_down(1);
	return;

    case DO_LEFT:	/* Move left (or stay at the margin )*/
	tty_cursor_left(1);
	return;

    case DO_RIGHT:	/* Move right (or stay at the margin )*/
	tty_cursor_right(1);
	return;

    case DO_DECI:	/* Index.  SET: Forward index; RESET: Back index */
	/* FIXME */
	break;

    case DO_SD:	/* Pan down Args[0] lines */
	/* FIXME */
	break;

    case DO_SU:	/* Pan up Args[0] lines */
	/* FIXME */
	break;

    case DO_CNL:	/* Do DO_NEL Args[0] times */
	/* FIXME */
	break;

    case DO_CPL:	/* Do DO_RI Args[0] times */
	for (i = 0; i < vt_ps->iArgs[0]; i++)
	    tty_cursor_up_scroll();
	return;

    case DO_CHA:	/* Move ANSI cursor to absolute column Args[0] */
	/* FIXME */
	break;

    case DO_CHI:	/* Cursor frwrd Args[0] TAB stops. Horizontal index */
	/* FIXME */
	break;


    case DO_CVA:	/* ANSI cursor to row Args[0] absolute */
	/* FIXME */
	break;

    case DO_NP:	/* Next page.  Args[0]: # of pages to move.
                                  New cursor position: Home */
	/* FIXME */
	break;

    case DO_PP:	/* Previous page.  Args[0]: # of pages to move.
                                  New cursor position: Home */
	/* FIXME */
	break;
    case DO_PPA:	/* Page position absolute.  Args[0]: page number. */
	/* FIXME */
	break;

    case DO_PPB:	/* Page position backward.  Args[0]: page number. */
	/* FIXME */
	break;

    case DO_PPR:	/* Page position relative.  Args[0]: page number. */
	/* FIXME */
	break;


    case DO_DECSC:	/* Save state: cursor position, graphic rendition,
                          character set shift state, wrap flag, origin mode,
                          state of selective erase */
	tty_save_state();
	return;

    case DO_DECRC:	/* Restores states described above. If none are saved,
                          cursor moves to the home position, origin more is
                          reset, no character attributes are assigned, the
                          default character set mapping is established */
	tty_restore_state();
	return;

    case DO_HTS:	/* Horizontal TAB set. Set a horizontal tab at the
                          current column */
	/* FIXME URGENT*/
	return;

    case DO_TBC:	/* Clear horizontal TAB stop at the current column */
	tty_set_tab_stop();
	break;

    case DO_TBCALL:	/* Clear all horizontal TAB stops */
	tty_clear_tab_stops();
	return;

    case DO_DECST8C:	/* Set TAB at every eight columns starting with
                          column 9 */
	tty_default_tab_stops();
	break;


    case DO_SGR:	/* Graphics Rendition. The command sets visual
                          attributes.  Args[0] = visual attribute.  For the
                          values of the attributes see vtconst.h ATTR_xxx.
                          Note: ATTR_NOT can be combined with all other
                          attributes (except ATTR_RESET). */
	set_attribute(vt_ps->iArgs[0]);
	return;

    case DO_SGRF:	/* Change foreground color Args[0]: color.
                          Bit values: 1=red, 2=green, 4=blue */
	tty_foreground_colour(vt_ps->iArgs[0] & 0x07);
	return;

    case DO_SGRB:	/* Change background color Args[0]: color.
                          Bit values: 1=red, 2=green, 4=blue */

	tty_background_colour(vt_ps->iArgs[0] & 0x07);
	return;

    case DO_DECDHL:	/* Double-height line. SET: Top half;
                                               RESET: Bottom half */
	/* FIXME */
	break;

    case DO_DECLW:	/* Line width.  SET: Double-width line;
                                        RESET: Single-width line */
	/* FIXME */
	break;

    case DO_DECHCCM:	/* Horizontal cursor-coupling mode.
                          SET: Coupled; RESET: Uncoupled. */
	/* FIXME */
	break;

    case DO_DECVCCM:	/* Vertical cursor-coupling mode.
                          SET: Coupled; RESET: Uncoupled. */
	/* FIXME */
	break;

    case DO_DECPCCM:	/* Page cursor-coupling mode.
                          SET: Coupled; RESET: Uncoupled. */
	/* FIXME */
	break;


    case DO_DECSCA:	/* Select Character Protection Attributes.
                          SET: the following characters are NOT erasable;
                          RESET: the following characters are erasable. */
	/* FIXME */
	break;

    case DO_IL:	/* Insert Line at the cursor position.
                          Args[0]: #lines */
	tty_insert_lines(vt_ps->iArgs[0]);
	return;

    case DO_DL:	/* Delete line at the cursor position.
                          Args[0]: #lines */
	tty_delete_lines(vt_ps->iArgs[0]);
	return;

    case DO_ICH:	/* Insert blank characters at the cursor position with
                          normal attributes.  Args[0]: #characters */
	tty_insert_chars(vt_ps->iArgs[0]);
	return;

    case DO_DCH:	/* Delete characters at the cursor position.
                          Args[0]: #characters */
	tty_delete_chars(vt_ps->iArgs[0]);
	return;

    case DO_DECIC:	/* Insert column starting with the cursor column.
                          Args[0]: number of columns to insert. */
	/* FIXME */
	break;

    case DO_DECDC:	/* Delete column starting with the cursor column.
                          Args[0]: number of columns to delete. */
	/* FIXME */
	break;

    case DO_ECH:	/* Erase characters at the cursor position.
                          Args[0]: #characters */
	/* FIXME */
	break;

    case DO_ELE:	/* Erase Line from the cursor position to the end of
                          the line. SET: erase all; RESET: only erasable */
	tty_clear_to_eol();
	return;

    case DO_ELB:	/* Erase Line from the beginning to the cursor
                          position. SET: erase all; RESET: only erasable */
	tty_clear_from_bol();
	return;

    case DO_EL:	/* Erase Line. SET: erase all; RESET: only erasable */
	/* FIXME */
	break;

    case DO_EDE:	/* Erase from the cursor position to the end of the
                          screen. SET: erase all; RESET: only erasable */
	tty_clear_to_eos();
	return;

    case DO_EDB:	/* Erase from the beginning of the screen to the cursor
                          position. SET: erase all; RESET: only erasable */
	/* FIXME */
	break;

    case DO_ED:	/* Erase Screen. SET: erase all;
                          RESET: only erasable */
	/* FIXME */
	tty_clear_screen();
	return;

    case DO_DECXRLM:	/* Transmit Rate Limiting. SET: limit transmit rate;
                          RESET: unlimited transmit rate */
	/* FIXME */
	break;

    case DO_DECHEM:	/* Hebrew Encoding Mode. SET: on; RESET: off */
	/* FIXME */
	break;

    case DO_DECRLM:	/* Cursor Right to Left Mode. SET: on; RESET: off */
	/* FIXME */
	break;

    case DO_DECNAKB:	/* Greek/N-A Keyboard Mapping. SET: Greek;
                          RESET: North American (note: the parser reverses
                          the native VT set/reset values for the sake of
                          consistency. )*/
	/* FIXME */
	break;

    case DO_DECHEBM:	/* Hebrew/N-A Keyboard Mapping. SET: Hebrew;
                          RESET: North American */
	/* FIXME */
	break;


    case DO_DECSTBM:	/* Set top and bottom margins for the scrolling
                         region. Args[0]: top; Args[1]: bottom. If Args[1]=1,
                         then the margin defaults to the bottom of the screen,
                         which is the current number of lines per screen. */
	tty_set_scroll_region(vt_ps->iArgs[0] - 1, vt_ps->iArgs[1] - 1);
	return;

    case DO_DECSLRM:	/* Set left and right margins. Args[0]: column # of
                          left margin; Args[1]: of right margin. If Args[1]=1,
                          right margin must be set to the right-most column
                          (80 or 132) depending on the page width. */
	/* FIXME */
	break;

    case DO_DECSCPP:	/* Set columns per page.  Args[0]: #columns */
	/* FIXME */
	break;

    case DO_DECSLPP:	/* Set the number of lines per page. Args[0]: #lines */
	/* FIXME */
	break;

    case DO_DECSNLS:	/* Set the number of lines per screen.
                          Args[0]: #lines */
	/* FIXME */
	break;


    case DO_AUTOPRN:	/* Auto Print mode. SET: ON; RESET: OFF */
	/* FIXME */
	break;

    case DO_PRNCTRL:	/* Printer Controller mode. SET: ON; RESET: OFF */
	/* FIXME */
	break;

    case DO_PRNCL:	/* Print the display line containing cursor
                          (and wait till done) */
	/* FIXME */
	break;

    case DO_PRNSCR:	/* Print screen (or the scrolling region)
                          (and wait till done) */
	/* FIXME */
	break;

    case DO_PRNHST:	/* Communication from the printer port to the host port
                          SET: Start; RESET: Stop */
	/* FIXME */
	break;

    case DO_PRNDSP:	/* Print Composed Display */
	/* FIXME */
	break;

    case DO_PRNALLPG:	/* Print All Pages */
	/* FIXME */
	break;

    case DO_ASSGNPRN:	/* Assign Printer to Active Host Session */
	/* FIXME */
	break;

    case DO_RELPRN:	/* Release Printer */
	/* FIXME */
	break;


    case DO_DA1:	/* "What is your service class code and what are
                          your attributes?" DA primary */
	/* FIXME */
	break;

    case DO_DA2:	/* "What terminal type are you, what version, and what
                          hardware options installed?" DA secondary */
	/* FIXME */
	break;

    case DO_DA3:	/* "What is your unit ID?" DA tertiary */
	/* FIXME */
	break;

    case DO_DSROS:	/* "Please report your operation status using DSR
                          control sequence" */
	/* FIXME */
	break;

    case DO_DSRCPR:	/* "Please report your cursor position using CPR
                          control sequence" */
	tty_send_cursor_position();
	return;

    case DO_DECEKBD:	/* Extended Cursor Position Report. */
	/* FIXME */
	break;

    case DO_DSRPRN:	/* "What is the printer status?" */
	/* FIXME */
	break;

    case DO_DSRUDK:	/* "Are user-defined keys locked or unlocked?" */
	/* FIXME */
	break;

    case DO_DSRKB:	/* "What is keyboard language?" */
	/* FIXME */
	break;

    case DO_DSRDECMSR:	/* Macro space report. */
	/* FIXME */
	break;

    case DO_DSRDECCKSR:	/* Memory checksum of text macro definitions.
                          Args[0] -> Pid (numerical label).  0 if none. */
	/* FIXME */
	break;

    case DO_DSRFLG:	/* "What is the status of your data integrity flag? */
	/* FIXME */
	break;

    case DO_DSRMULT:	/* "What is the status of the multiple-session
                          configuration?" */
	/* FIXME */
	break;

    case DO_DECRQCRA:	/* Request a checksum of the rectangular area.
                          Args[0] = numeric Pid
                          Args[1] = page number.  If Args[1] = 0, return a
                            checksum for all pages ignoring all parameters.
                          Args[2] = top-line border
                          Args[3] = left-column border
                          Args[4] = bottom-line border
                            Args[4] = 0 -> last line of the page
                          Args[5] = right-column border
                            Args[5] = 0 -> last column of the page */
	/* FIXME */
	break;

    case DO_DECRQSTAT:	/* Request entire machine state */
	/* FIXME */
	break;

    case DO_DECRQTSR:	/* Request Terminal State Report */
	/* FIXME */
	break;

    case DO_DECRQTSRCP:	/* Request color palette */
	/* FIXME */
	break;

    case DO_DECRQUPSS:	/* Request User Preferred Supplemental Set */
	/* FIXME */
	break;

    case DO_DECRQCIR:	/* Request cursor information report */
	/* FIXME */
	break;

    case DO_DECRQTAB:	/* Request Tab stop report */
	/* FIXME */
	break;

    case DO_DECRQMANSI:	/* Request ANSI mode controls.
                             Args[0]: mode (see RQM_ vtconst.h) */
	/* FIXME */
	break;

    case DO_DECRQMDEC:	/* Request DEC mode controls.
                             Args[0]: mode (see RQM_ vtconst.h) */
	/* FIXME */
	break;

    case DO_DECREQTPARM:	/* Request terminal parameters
                             SET: generate DECREPTPARMs only in response to a
                               request;
                             RESET: can send unsolicited DECREPTPARMs. */
	/* FIXME */
	break;

    case DO_DECRQDE:	/* Request Displayed Extent */
	/* FIXME */
	break;


    case DO_DECSTR:	/* Soft Reset.  Selects most of the power-up factory-
                            default settings */
	/* FIXME */
	break;

    case DO_RIS:	/* Hard Reset.  Selects the saved settings saved in
                            nonvolatile memory */
	/* FIXME */
	break;

    case DO_DECSR:	/* Secure Reset.  Sets the terminal to its power-up
                            state to guarantee the terminal state for secure
                            connections.
                              Args[0] = secure reset id
                                      = -1 no id; do not send DECSRC */
	/* FIXME */
	break;


    case DO_DECTSTMLT:	/* Do: DECTSTxxx, xxx are POW, RSP, PRN, BAR, RSM */
	/* FIXME */
	break;

    case DO_DECTSTPOW:	/* Power-up self-test */
	/* FIXME */
	break;

    case DO_DECTSTRSP:	/* RS-232 port loopback test */
	/* FIXME */
	break;

    case DO_DECTSTPRN:	/* Printer port loopback test */
	/* FIXME */
	break;

    case DO_DECTSTBAR:	/* Color bar test */
	/* FIXME */
	break;

    case DO_DECTSTRSM:	/* Modem control line loopback test */
	/* FIXME */
	break;

    case DO_DECTST20:	/* 20 mA port/DEC-420 loopback test */
	/* FIXME */
	break;

    case DO_DECTSTBLU:	/* Full screen blue */
	/* FIXME */
	break;

    case DO_DECTSTRED:	/* Full screen red */
	/* FIXME */
	break;

    case DO_DECTSTGRE:	/* Full screen green */
	/* FIXME */
	break;

    case DO_DECTSTWHI:	/* Full screen white */
	/* FIXME */
	break;

    case DO_DECTSTMA:	/* Internal modem analog loopback test */
	/* FIXME */
	break;

    case DO_DECTSTME:	/* Internal modem external loopback test */
	/* FIXME */
	break;

    case DO_DECALN:	/* Display full screen of E's */
	/* FIXME */
	break;


    case DO_VTMODE:	/* Change VTxxx mode. Args[0]: VT_MODE52, VT_MODE100,
                          VT_MODE420. Args[1]: (for VT420) VT_7BIT, VT_8BIT.
                          VT52 and VT100 are always 7 bit) */
	/* FIXME */
	return;

    case DO_ANSICONF:	/* Change ANSI conformance level
                          Args[0] - desired level (1, 2, or 3). */
	/* FIXME */
	break;

    case DO_DECPCTERM:	/* Change PC Emulation mode.
                          Args[0] - operating mode
                                  = 0 -> VT mode
                                  = 1 -> PC TERM mode
                          Args[1] = PC character set
                                   (see vtconst.h for PCCS_xxx) */
	/* FIXME - ignore this right now*/
	return;


    case DO_DECLL:	/* LED. RESET: All LED off. Otherwise, turn LED on,
                          where Args[0] is LED# (usually, 1-4) */
	/* FIXME */
	break;

    case DO_DECSSDT:	/* Select Status line type.
                          Args[0] = 0 -> No status line;
                          Args[0] = 1 -> Local owned (indicator line);
                          Args[0] = 2 -> Host-owned (remote status line) */
	/* FIXME */
	break;


    case DO_DECES:	/* Enable Session */
	/* FIXME */
	break;


    case DO_DECCRA:	/* Copy rectangular area.
                          Args[0] = SOURCE: top-line border
                          Args[1] = SOURCE: left-column border
                          Args[2] = SOURCE: bottom-line border
                            Args[2] = 0 -> last line of the page
                          Args[3] = SOURCE: right-column border
                            Args[3] = 0 -> last column of the page
                          Args[4] = SOURCE: page number
                          Args[5] = DESTINATION: top-line border
                          Args[6] = DESTINATION: left-column border
                          Args[7] = DESTINATION: page number */
	/* FIXME */
	break;

    case DO_DECERA:	/* Erase rectangular area.
                          Args[0] = top-line border
                          Args[1] = left-column border
                          Args[2] = bottom-line border
                            Args[2] = 0 -> last line of the page
                          Args[3] = right-column border
                            Args[3] = 0 -> last column of the page */
	/* FIXME */
	break;

    case DO_DECFRA:	/* Fill rectangular area.
                          Args[0] = fill character
                          Args[1] = top-line border
                          Args[2] = left-column border
                          Args[3] = bottom-line border
                            Args[3] = 0 -> last line of the page
                          Args[4] = right-column border
                            Args[4] = 0 -> last column of the page */
	/* FIXME */
	break;

    case DO_DECSERA:	/* Selective erase rectangular area.
                          Args[0] = top-line border
                          Args[1] = left-column border
                          Args[2] = bottom-line border
                            Args[2] = 0 -> last line of the page
                          Args[3] = right-column border
                            Args[3] = 0 -> last column of the page */
	/* FIXME */
	break;

    case DO_DECSACE:	/* Select attribute change extent.
                          SET: change stream of character positions
                          RESET: change rect area of character positions */
	/* FIXME */
	break;

    case DO_DECCARA:	/* Change visual attributes in rectangular area.
                          Args[0] = top-line border
                          Args[1] = left-column border
                          Args[2] = bottom-line border
                            Args[2] = 0 -> last line of the page
                          Args[3] = right-column border
                            Args[3] = 0 -> last column of the page
                          Args[4] = attributes to SET
                          Args[5] = attributes to CLEAR
                             (See ATTR_xxx in vtconst.h)
                             Note: the attributes may be cleared and set in
                               any order since there will be NO attributes
                               set in BOTH Args[4] and Args[5].  The attributes
                               to be set or cleared are Bold, Underline,
                               Blinking, Negative, and Invisible.
                          Args[n+1]..Args[VTARGS-1] = ATTR_NONE (-1) */
	/* FIXME */
	break;

    case DO_DECRARA:	/* Reverse visual attributes in rectangular area.
                          Args[0] = top-line border
                          Args[1] = left-column border
                          Args[2] = bottom-line border
                            Args[2] = 0 -> last line of the page
                          Args[3] = right-column border
                            Args[3] = 0 -> last column of the page
                          Args[4] = attributes to be reversed.  The attributes
                            that can be reversed are Bold, Underline, Blinking,
                            Negative, and Invisible. */
	/* FIXME */
	break;


    case DO_DECELF:	/* Enable (and disable) Local Functions.
                          Args[0] = 0 -> all local function
                                    1 -> local copy and paste
                                    2 -> local panning
                                    3 -> local window resize
                          Args[1] = 0 -> factory default
                                    1 -> enable
                                    2 -> disable */
	/* FIXME */
	break;

    case DO_DECLFKC:	/* Local Functions Key Control.
                          Args[0] = 0 -> all local function keys
                                    1 -> F1 [Hold]
                                    2 -> F2 [Print]
                                    3 -> F3 [Set-Up]
                                    4 -> F4 [Session]
                          Args[1] = 0 -> factory default
                                    1 -> local function
                                    2 -> send key sequence
                                    3 -> disable */
	/* FIXME */
	break;

    case DO_DECSMKR:	/* Select Modifier Key Reporting.
                          Args[0] = 0 -> all keys
                                    1 -> left [Shift]
                                    2 -> right [Shift]
                                    3 -> [Caps Lock]
                                    4 -> [Ctrl]
                                    5 -> left [Alt Function]
                                    6 -> right [Alt Function]
                                    7 -> left [Compose Character]
                                    8 -> right [Compose Character]
                          Args[1] = 0 -> factory default
                                    1 -> modifier function
                                    2 -> extended keyboard report
                                    3 -> disabled */
	/* FIXME */
	break;


  /* The following functions are the set of Locator Extentions implemented in
   several DEC terminals.  These are often unsupported by other parsers/
   emulators */

    case DO_DECELR:	/* Enable Locator Reports.
                          Args[0] = 0 -> locator disabled
                                    1 -> locator reports enabled
                                    2 -> one shot (allow one report and
                                           disable)
                          Args[1] - specifies coordinate units
                                  = 0 or 2 -> character cells
                                    1 -> device physical pixels */
	/* FIXME */
	break;

    case DO_DECSLE:	/* Select Locator Events.
                          Args[0] = 0 -> report only to explicit host requests
                                    1 -> report button down transitions
                                    2 -> do not report button down transitions
                                    3 -> report button up transitions
                                    4 -> do not report button up transitions */
	/* FIXME */
	break;

    case DO_DECRLP:	/* Request Locator Position. */
	/* FIXME */
	break;

    case DO_DECEFR:	/* Enable Filter Rectangle.
                          Args[0] = Top boundary of filter rectangle
                          Args[1] = Left boundary of filter rectangle
                          Args[2] = Bottom boundary of filter rectangle
                          Args[3] = Right boundary of filter rectangle
                          NOTE: The origin is 1,1.  If any of the arguments are
                            equal to 0, they default to the current locator
                            coordinate. */
	/* FIXME */
	break;

    case DO_DSRLOCATOR:	/* Request locator device status */
	/* FIXME */
	break;

    case DO_LOCATORCTRL:	/* Locator controller mode. SET: ON; RESET: OFF */
	/* FIXME */
	break;


    case DO_THEEND:	/* This command is called for the termination of
                                the command, specified with PA_ANNOUNCED
                                in "vtcmnd.h" .  Use it when yu need to know
                                when the last argument in some multi-argument
                                string has been passed.
                                Example: DO_DECRQMANSI */
	/* FIXME */
	break;

    case DO_HOLD:	/* This command is called repeatedly while parser
								is is a HOLD state.  To allow parsing to
								continue, HOLD state must be reset by the code
								that processes this command */
	/* FIXME */
	break;


    case DO_DECUDK:	/* Set User Definable Keys.
                          First, the command is called with:
                            Args[0] = 0 -> clear all UDK definitions.
                                      1 -> clear only keys to be defined
                            Args[1] = 0 -> lock the keys
                                      1 -> do not lock the keys
                            Args[2] = 1 -> define unshifted function key
                                      2 -> define shifted function key
                                      3 -> define alternate unshifted F-key
                                      4 -> define alternate shifted F-key

                          After, while loading UDK,
                            Args[0] = -1
                            Args[1] = DEC F key number defined: (1-20)
                              Note: if Args[1] < 0, then this DEC F key number
                              is -Args[1] and is an ALTERNATE (no matter what
                              Args[1] in the initial call claimed, which,
                              however, still defines whether the key is
                              shifted alternate or an unshifted alternate).
                            Args[2]:Args[3] = integer encoded pointer to a
                              NULL-terminating string of ASCII codes
                              assigned to the key (use INTSTOPOINTER).
                            Args[4] = 0 -> errors in string.  attempted to
                                             correct.
                                      1 -> ok.  no errors.

                          When the sequence ended:
                            Args[0] = Args[1] = -1

                            Args[2] = 0 -> abnormal termination (CAN, etc. )
                                      1 -> normal termination (with ST) */
	/* FIXME */
	break;

    case DO_DECDLD:	/* Downline Load soft character set.
                          First, the command is called with:
                            Args[0] = 0 or 1. Font number (buffer number).
                                                Both result in DRCS 1 in VT420.
                            Args[1] = First character to load.
                                       Range of Args[1]: 0x20 - 0x7F.
                            Args[2] = 0 -> erase all chars in DRCS with this
                                             number, width, and rendition
                                      1 -> erase only chars in locations being
                                             reloaded
                                      2 -> erase all renditions of the char set
                               Note: erased chars are undefined (not black).
                                     They shall be displayed as the error char.
                            Args[3] = 2 -> 5x10 pixel cell (V220 compatible)
                                      3 -> 6x10 pixel cell (V220 compatible)
                                      4 -> 7x10 pixel cell (V220 compatible)
                                      other -> Args[3] = pixel width (5 - 10)
                            Args[4] = #columns and lines.
                                        24 + 12 * (Args[4] % 16) --> #lines
                                        80 + 52 * (Args[4] / 16) --> #columns
                            Args[5] = 1 -> text font (do spacing and centering)
                                      2 -> full cell (display as is)
                            Args[6] = pixel height (# of pixels) (1 - 16)
                            Args[7] = 0 -> 94-character set
                                      1 -> 96-character set
                            Args[8] - final char (see CS_xx in vt_const.h for
                              the final chars for the standard hard char sets)
                            Args[9] - encoded intermediate chars.  Formula:
                               Args[9] = #_of_intermediates * 0x0100 +
                                        (intermediate_#_1 - ' ') * 0x0010 +
                                        (intermediate_#_2 - ' ')
                                NOTE: You do NOT need to decode this
                                      argument since the parser is using the
                                      same encoding formula in all functions. */

			/*
                          After, while loading DLD,
                            Args[0] = -1
                            Args[1] = # of the char loaded (1-first, 2-second,
                              3-third, ...)  This argument will probably have
                              little use, but...
                            Args[2]:Args[3] = integer encoded pointer to a
                              NULL-terminating string of ASCII sixels that
                              define the next character.  The sixel groups are
                              separated by "/" char, just as received in the
                              first place (with 0x3F offset still in place)
                              (use INTSTOPOINTER).

                          When the sequence ended:
                            Args[0] = Args[1] = -1
                            Args[2] = 1 -> normal termination (with ST)
                                      0 -> abnormal termination (CAN, etc. )
                            Args[3] = total # of chars loaded. */
	/* FIXME */
	break;


    case DO_DECRQSS:	/* Request control function setting.  The response
                          expected is the command that would do that function
                          except the leading CSI is omitted.

                          Args[0] is the command that is requested.  Standard
                            VT420 commands that could be requested this way
                            are: DO_DECELF, DO_DECLFKC, DO_DECSASD, DO_DECSACE,
                            DO_SGR, DO_DECSMKR, DO_DECSCA, DO_DECSCPP,
                            DO_VTMODE, DO_DECSLRM, DO_DECSLPP, DO_DECSNLS,
                            DO_DECSSDT, DO_DECSTBM.  However, the parser will
                            lookup any command in vtcsi.tbl table.  If a
                            command is not found, DO_VTERR is returned.
                          Args[1]..Args[VTARGS-1] contain the ACSII chars
                            passed in DECRQSS string, which were used to
                            identify the command, and will have to be echoed
                            back (with the other info).  The unused places are
                            filled with -1.

                          Note: the command in Args[0] is either the only
                            command or the default command associated with the
                            given char identifier and flags. */
	/* FIXME */
	break;


    case DO_DECDMAC:	/* Define macro.
                            Args[0] = macro ID number.  range: 0 and 63.
                            Args[1] = 0 -> delete only the macro defined
                                      1 -> delete all current macro definitions
                            Args[2]:Args[3] = integer encoded pointer to a
                              NULL-terminating string of ASCII characters that
                              define the macro (use INTSTOPOINTER).

                            Args[4] = 0 -> errors in macro. attempted to
                                             correct.
                                      1 -> ok.  no errors in macro.

                            Args[5] = 0 -> abnormal termination (CAN, etc. )
                                      1 -> normal termination (with ST) */
	/* FIXME */
	break;


    case DO_DECINVM:	/* Invoke macro.  Args[0] = macro ID number */
	/* FIXME */
	break;


    case DO_DECAUPSS:	/* Assign User-Preferred Character Set
                            Args[0] - final char (see CS_xx in vt_const.h for
                              the final chars for the standard hard char sets)
                            Args[1] - encoded intermediate chars.  Formula:
                              Args[1] = #_of_intermediates * 0x0100 +
                                       (intermediate_#_1 - ' ') * 0x0010 +
                                       (intermediate_#_2 - ' ')
                              NOTE: You do NOT need to decode this
                                    argument since the parser is using the
                                    same encoding formula in all functions. */
	/* FIXME */
	break;


    case DO_DECRSTS:	/* Restore Terminal to a previous state specified in a
                          terminal state report (DECTSR).
                            Args[0]:Args[1] -> integer encoded pointer to the
                                              NULL-terminated data string
                                              (use INTSTOPOINTER).
                            Args[2] = 0 -> abnormal termination (CAN, etc. )
                                      1 -> normal termination (with ST)

                           Note: RSTSLEN parameter in vt420.h defines the size
                             of the buffer (the actual space allocated is
                             RSTSLEN + 1 to accommodate for the null character
                             at the end of the string. */
	/* FIXME */
	break;


    case DO_DECRSCI:
			/* Restore Cursor Information.  The arguments are
                          passed one-by-one.
                            Args[0] = number of the data element reported
                              (the first data element has the number 1).
                              The DEC mnemonics for the elements are:
                                       1  -> Pr
                                       2  -> Pc
                                       3  -> Pp
                                       4  -> Srend
                                       5  -> Sarr
                                       6  -> Sflag
                                       7  -> Pgl
                                       8  -> Pgr
                                       9  -> Scss
                                       10 -> Sdesig
                            Args[1] = 0 -> error in the argument.  attempted
                                             to correct
                                      1 -> ok.  no error
                            All Pxx elements (or 0 if omitted) are returned
                              in Args[2].
                            All Sxxx except Sdesig are returned as
                              Args[2]:Args[3] as integer encoded pointer to a
                                    null-terminating string containing
                                    the parameter.
                              Sdesig is returned as:
                                Args[2] - final char for G0
                                Args[3] - encoded intermediate chars for G0
                                Args[4] - final char for G1
                                Args[5] - encoded intermediate chars for G1
                                Args[6] - final char for G2
                                Args[7] - encoded intermediate chars for G2
                                Args[8] - final char for G3
                                Args[9] - encoded intermediate chars for G3

                                A final char that is equal to -1 indicates an
                                  error in this Cdesig argument.

                                See CS_xx in vt_const.h for the final chars
                                for the standard hard char sets)
                                The formula for encoded intermediate chars is
                                  Args[x] = #_of_intermediates * 0x0100 +
                                        (intermediate_#_1 - ' ') * 0x0010 +
                                        (intermediate_#_2 - ' ')
                                NOTE: You do NOT need to decode this
                                      argument since the parser is using the
                                      same encoding formula in all functions.
                                HOWEVER: to use DECCIR, you will need to use
                                  PA_DECODEINTERMS macro.

                            When DECRSCI data is over
                              Args[0] = -1
                              Args[1] = 0 -> abnormal termination (CAN, etc. )
                                        1 -> normal termination (with ST or too
                                            many parameters)
                              Args[2] = 0 -> the data string is too long (too
                                            many params).  Extras are ignored.
                                        1 -> either enough or too few params
                                            then needed.  See Args[3] to find
                                            out which.
                              Args[3] = total number of params encountered */
	/* FIXME */
	break;


    case DO_DECRSTAB:	/* Restore TAB Information.
                          Args[0] = tab position or 0 is omitted
                          Args[1] = 0 -> error in data.  attempted to correct
                                    1 -> ok.  no error

                            NOTE: If the DCS string does not specify ANY TAB
                              information, one call to DO_DECRSTAB with
                              Args[0] = 0 will be made.

                          When DECRSTAB data is over,
                            Args[0] = -1
                            Args[1] = 0 -> abnormal termination (CAN, etc. )
                                      1 -> normal termination (with ST) */

    /* The following group of functions are NOT VT standard ones.  However, a
      few terminal programs do understand them.  Please note that while
      XTRANS, XRECEIVE, XAPPEND, and XSAVE _are_ used by some terminal
      emulators, XSUPP and XOK / XERROR  (XOK is 'ESC { + '  and XERROR is
      'ESC { - ') are unique to this parser implementation.  */
	/* FIXME */
	break;


    case DO_XSUPP:	/* "Do you support the extended functions?"  If you
                          do support the following functions, respond with
                                            ESC { +
                          Otherwise, ignore */
	/* FIXME */
	break;

    case DO_XTRANS:	/* Transmit a file.
                          Args[0]:Args[1] - integer encoded pointer to a null-
                            terminated string with the filename
                            (use INTSTOPOINTER).
                          If the application agrees to honor the request,
                          respond with " ESC { + ".  If there is an error,
                          respond with " ESC { - ".  If you do not support
                          extended functions, ignore the request. */
	/* FIXME */
	break;

    case DO_XRECEIVE:	/* Receive a file.
                          Args[0]:Args[1] - integer encoded pointer to a null-
                            terminated string with the filename
                            (use INTSTOPOINTER).
                          If the application agrees to honor the request,
                          respond with " ESC { + ".  If there is an error,
                          respond with " ESC { - ".  If you do not support
                          extended functions, ignore the request. */
	/* FIXME */
	break;

    case DO_XAPPEND:	/* Append to a file.
                          Args[0]:Args[1] - integer encoded pointer to a null-
                            terminated string with the filename
                            (use INTSTOPOINTER).
                          If the application agrees to honor the request,
                          respond with " ESC { + ".  If there is an error,
                          respond with " ESC { - ".  If you do not support
                          extended functions, ignore the request. */
	/* FIXME */
	break;

    case DO_XSAVE:	/* Save a file collected text.  This command is used
                          with XRECEIVE and XAPPEND.
                          If the application agrees to honor the request,
                          respond with " ESC { + ".  If there is an error,
                          respond with " ESC { - ".  If you do not support
                          extended functions, ignore the request. */
	/* FIXME */
	break;

    }
    error_message("<<%d>>", code);
    return;
}

