
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <windows.h>
#include <winsock.h>
#include "vt.h"
#include "emulator.h"

#define	IAC	255		/* interpret as command: */
#define	DONT	254		/* you are not to use option */
#define	DO	253		/* please use option */
#define	WONT	252		/* I won't use option */
#define	WILL	251		/* I will use option */
#define SB	250		/* interpret as subnegotiation */
#define SE	240		/* end sub negotiation */
#define	TELOPT_TTYPE	24
#define	TELOPT_NAWS	31	/* window size */
#define	TELOPT_LFLOW	33	/* remote flow control */
#define	TELOPT_LINEMODE	34	/* Linemode option */
#define	TELOPT_XDISPLOC	35
#define	TELOPT_ECHO	1	/* echo */
#define	TELOPT_NEW_ENVIRON 39	/* New - Environment variables */

#define TELQUAL_IS      0       /* option is... */
#define TELQUAL_SEND    1       /* send option */
#define TELQUAL_INFO    2       /* ENVIRON: informational version of IS */
#define TELQUAL_REPLY   2       /* AUTHENTICATION: client version of IS */
#define TELQUAL_NAME    3       /* AUTHENTICATION: client version of IS */

#define NEW_ENV_VAR     0
#define NEW_ENV_VALUE   1
#define ENV_ESC         2
#define ENV_USERVAR     3

/*
** Maximum allowable screen size
*/
#define MAX_SCREEN_WIDTH	200
#define MAX_SCREEN_HEIGHT	120

/*
** Number of pixels around text region
*/
#define SCREEN_BORDER		2


BOOL	mode_echo = FALSE;
BOOL	mode_insert = FALSE;
BOOL	mode_fix_window_size = TRUE;
char	*tty_fontname = "Lucida Console";

extern HWND	hwndMain;

static int	screen_width = 80;
static int	screen_height = 24;

static HDC	hScreenDC;
static HBITMAP	hScreenBitmap;

char	*terminal_type = "VT220";

typedef long t_char;

typedef struct t_line
{
    /*
    ** Has this line been modified?
    */
    BOOL	modified;

    /*
    ** Number of blinking characters on this line
    */
    int		blinks;

    t_char	data[MAX_SCREEN_WIDTH];
} t_line;

/*
** Contents of the screen
*/
static t_char	screen_contents[MAX_SCREEN_WIDTH][MAX_SCREEN_HEIGHT + 1];

/*
** Contents of the current screen buffer
*/
static t_line	**screen_output;
static t_line	**screen_display;
static int	screen_viewport_top = 0;

/*
**
*/
static t_line	**screen_history = NULL;
int		max_screen_history = 0;

/*
** Tab stops
*/
static int tab_stops[MAX_SCREEN_WIDTH];

/*
** Spacing between characters
*/
static INT	font_spacing[MAX_SCREEN_WIDTH];

/*
** On vt terminal, when the last character of a row is outputted, the cursor is considered
** to be still on the same line...
*/
static BOOL	vt_cursor_at_eol = FALSE;

/*
** Requested position of cursor
*/
static int	cursor_on = TRUE;
static int	cursor_x = 0;
static int	cursor_y = 0;

/*
** Font use for display text (must be fixed width font)
*/
static HFONT	font;
static HFONT	font_bold;

/*
** Width and height of font
*/
int	font_height = 16;
int	font_width = 0;


/*
** Top and bottom line of scrolling region
*/
static int	scroll_top = 0;
static int	scroll_bottom = 0;

/*
** Size of font divided by 2
*/
static int	half_height;
static int	half_width;

/*
** Rectangle containing the screen area
*/
static RECT	terminal_rect;

/*
** Current attribute value character would take when
** output to screen
*/
static t_char 	curr_attribute;
static BOOL	blink_shown = TRUE;

static int	curr_foreground_colour = -1;
static int	curr_background_colour = -1;

COLORREF	colour_background = RGB(0, 0, 0);
COLORREF	colour_text = RGB(192, 192, 192);
COLORREF	colour_bold = RGB(255, 255, 255);
COLORREF	colour_cursor = RGB(192, 192, 192);
COLORREF	colour_cursor_text = RGB(0, 0, 0);


static HPEN	pen_text;
static HPEN	pen_bold;
static HPEN	pen_cursor_text;
static HPEN	pen_background;
static HBRUSH	brush_text;
static HBRUSH	brush_bold;
static HBRUSH	brush_background;
static HBRUSH	brush_cursor;

#define MAX_TTY_COLOUR	16
static HBRUSH	brush_colours[MAX_TTY_COLOUR];
static HPEN	pen_colours[MAX_TTY_COLOUR];
static COLORREF	tty_colours[MAX_TTY_COLOUR] =
{
    /* Normal intensity colour */
    RGB(0, 0, 0),
    RGB(128, 0, 0),
    RGB(0, 128, 0),
    RGB(128, 128, 0),
    RGB(0, 0, 128),
    RGB(128, 0, 128),
    RGB(0, 128, 128),
    RGB(192, 192, 192),

    /* Bold colour */
    RGB(0, 0, 0),
    RGB(255, 0, 0),
    RGB(0, 255, 0),
    RGB(255, 255, 0),
    RGB(0, 0, 255),
    RGB(255, 0, 255),
    RGB(0, 255, 255),
    RGB(255, 255, 255)
};

static HPALETTE	palette;

static void do_move_cursor(int, int);

/*
** Construct the string for sending window size to the telnet server
*/
static char	*send_tty_size(char *p)
{
    *p++ = IAC;
    *p++ = SB;
    *p++ = TELOPT_NAWS;
    *p++ = (screen_width / 256);
    *p++ = (screen_width % 256);
    *p++ = (screen_height / 256);
    *p++ = (screen_height % 256);
    *p++ = IAC;
    *p++ = SE;
    return p;
}

static void	cursor_hide(BOOL hide)
{
    t_line	*line;

    if (!cursor_on)
	return;

    line = screen_output[cursor_y];

    if (hide)
	line->data[cursor_x] &= ~ZATTR_CURSOR;
    else
	line->data[cursor_x] |= ZATTR_CURSOR;

    line->modified = TRUE;
}

static void	clear_line(t_line *line)
{
    int		x;

    line->blinks = 0;
    line->modified = TRUE;

    for (x = 0; x < screen_width; x++)
	line->data[x] = (curr_attribute & ZATTR_COLOUR_MASK) | ' ';
}

void	tty_set_size(int cols, int rows)
{
    screen_width = cols;
    screen_height = rows;
}

static void	tty_resize_screen(int new_width, int new_height)
{
    int		i, j;

    cursor_hide(TRUE);

    if (new_height < screen_height && cursor_y >= new_height)
    {
	while (cursor_y > 0 && cursor_y >= new_height)
	{
	    t_line	*line = screen_output[0];

	    for (j = 0; j < MAX_SCREEN_HEIGHT; j++)
	    {
		screen_output[j] = screen_output[j + 1];
		screen_output[j]->modified = TRUE;
	    }
	    screen_output[j] = line;
	    screen_output[j]->modified = TRUE;
	    cursor_y--;
	}
    }

    for (i = 0; i < new_height + 1; i++)
    {
	t_line	*line = screen_output[i];

	line->blinks = 0;
	line->modified = TRUE;
	for (j = 0; j < new_width; j++)
	    if (line->data[j] & ZATTR_BLINK)
		line->blinks++;
    }

    screen_width = new_width;
    screen_height = new_height;
    tty_set_scroll_region(0, 0);
    do_move_cursor(cursor_x, cursor_y);
}

static void	set_palette_entry(PALETTEENTRY *p, COLORREF c)
{
    p->peRed = GetRValue(c);
    p->peGreen = GetGValue(c);
    p->peBlue = GetBValue(c);
    p->peFlags = 0;
}

static	void	*zmalloc(int size)
{
    void	*p = malloc(size);

    memset(p, 0, size);
    return p;
}

void	tty_startup(void)
{
    static int	started = FALSE;
    int		x, y;
    int		i;
    char	buf[128];
    char	*p;
    HDC		hdc;
    int		n_lines = max_screen_history + MAX_SCREEN_HEIGHT + 1;

    if (started)
	return;

    started = TRUE;

    screen_history = zmalloc(sizeof(t_line *) * n_lines);
    for (y = 0; y < n_lines; y++)
    {
	t_line	*line = zmalloc(sizeof(t_line));

	for (x = 0; x < MAX_SCREEN_WIDTH; x++)
	    line->data[x] = ' ';

	screen_history[y] = line;
    }

    /*
    ** Set the viewport to the first screen lines after the history
    */
    screen_viewport_top = max_screen_history;

    screen_output = &screen_history[max_screen_history];
    screen_display = &screen_history[max_screen_history];

    hdc = GetDC(hwndMain);
    if (GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE)
    {
	LOGPALETTE	*lp = (LOGPALETTE *)malloc(sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * 6));

	lp->palVersion = 0x300;
	lp->palNumEntries = 5;

	set_palette_entry(&lp->palPalEntry[0],  colour_background);
	set_palette_entry(&lp->palPalEntry[1],  colour_text);
	set_palette_entry(&lp->palPalEntry[2],  colour_bold);
	set_palette_entry(&lp->palPalEntry[3],  colour_cursor);
	set_palette_entry(&lp->palPalEntry[4],  colour_cursor_text);

	palette = CreatePalette(lp);

	if (palette)
	{
	    colour_background = PALETTEINDEX(0);
	    colour_text = PALETTEINDEX(1);
	    colour_bold = PALETTEINDEX(2);
	    colour_cursor = PALETTEINDEX(3);
	    colour_cursor_text = PALETTEINDEX(4);
	}

    }
    ReleaseDC(hwndMain, hdc);


    pen_text = CreatePen(PS_SOLID, 1, colour_text);
    pen_bold = CreatePen(PS_SOLID, 1, colour_bold);
    pen_cursor_text = CreatePen(PS_SOLID, 1, colour_cursor_text);
    pen_background = CreatePen(PS_SOLID, 1, colour_background);

    brush_text = CreateSolidBrush(colour_text);
    brush_bold = CreateSolidBrush(colour_bold);
    brush_background = CreateSolidBrush(colour_background);
    brush_cursor = CreateSolidBrush(colour_cursor);

    for (i = 0; i < MAX_TTY_COLOUR; i++)
    {
	pen_colours[i] = CreatePen(PS_SOLID, 1, tty_colours[i]);
	brush_colours[i] = CreateSolidBrush(tty_colours[i]);
    }

    p = buf;
    *p++ = IAC;
    *p++ = WONT;
    *p++ = TELOPT_LINEMODE;

    tty_set_scroll_region(0, 0);

    port_write_str(buf, p - buf);

}

void	tty_set_scroll_region(int top, int bottom)
{
    if (bottom <= 0)
	bottom = screen_height - 1;
    else if (bottom > screen_height - 1)
	bottom = screen_height - 1;

    if (top < 0)
	top = 0;
    else if (top > screen_height - 1)
	top = screen_height - 1;

    scroll_top = top;
    scroll_bottom = bottom;
}

void	tty_scroll_down(void)
{
    int		y;
    t_line	*l1 = screen_output[scroll_bottom];

    cursor_hide(TRUE);

    for (y = scroll_bottom; y > scroll_top; y--)
    {
	screen_output[y] = screen_output[y - 1];
	screen_output[y]->modified = TRUE;
    }

    screen_output[y] = l1;

    clear_line(l1);

    cursor_hide(FALSE);
}


void	tty_scroll_up(void)
{
    int		y;
    t_line	*spare;

    cursor_hide(TRUE);

    if (scroll_top == 0)
    {
	spare = screen_history[0];
	for (y = 0; y < max_screen_history; y++)
	{
	    screen_history[y] = screen_history[y + 1];
	    screen_history[y]->modified = TRUE;
	}
    }
    else
    {
	spare = screen_output[scroll_top];
    }

    for (y = scroll_top; y < scroll_bottom; y++)
    {
	screen_output[y] = screen_output[y + 1];
	screen_output[y]->modified = TRUE;
    }

    screen_output[y] = spare;

    clear_line(spare);

    cursor_hide(FALSE);
}

void	tty_cursor_on(void)
{
    if (!cursor_on)
    {
	cursor_on = TRUE;
	cursor_hide(FALSE);
    }
}

void	tty_cursor_off(void)
{
    if (cursor_on)
    {
	cursor_hide(TRUE);
	cursor_on = FALSE;
    }
}

void	tty_echo(int ch)
{
    if (mode_echo)
	tty_output(ch);
}

#define DEFAULT_FONT	"Courier New"

void	set_palette(HDC hdc)
{
    if (palette)
    {
	SelectPalette(hdc, palette, FALSE);
	RealizePalette(hdc);
    }
}

void	tty_init(HWND hWnd)
{
    LOGFONT	lf;
    TEXTMETRIC	tm;
    HDC		hdc;
    RECT	client;
    int		dx, dy;
    int		x, y;
    int		retry;
    char	facename[80];

    if (screen_history == NULL)
    	return;

    hdc = GetDC(hWnd);
    set_palette(hdc);
    GetClientRect(hWnd, &client);


    if (hScreenDC == NULL)
	DeleteDC(hScreenDC);
    if (hScreenBitmap != NULL)
	DeleteObject(hScreenBitmap);

    hScreenBitmap = CreateCompatibleBitmap(hdc, client.right - client.left, client.bottom - client.top);
    hScreenDC = CreateCompatibleDC(hdc);
    SelectObject(hScreenDC, hScreenBitmap);
    set_palette(hScreenDC);

    FillRect(hScreenDC, &client, brush_background);
    for (y = 0; y < screen_height; y++)
    {
	t_line	*line = screen_display[y];
	line->modified = TRUE;
	line->blinks = 0;

	for (x = 0; x < screen_width; x++)
	{
	    if (line->data[x] & ZATTR_BLINK)
		line->blinks++;
	    screen_contents[x][y] = ' ';
	}
    }


    InflateRect(&client, -SCREEN_BORDER, -SCREEN_BORDER);

    memset(&lf, 0, sizeof(LOGFONT));

    if (!mode_fix_window_size)
    {
	lf.lfWidth = font_width;
	lf.lfHeight = font_height;
    }
    else
    {
	lf.lfWidth = (client.right - client.left) / (screen_width);
	lf.lfHeight = (client.bottom - client.top) / (screen_height);
    }

    lf.lfHeight = max(4, lf.lfHeight);

    strcpy(lf.lfFaceName, tty_fontname);

    if (mode_fix_window_size || !font)
    {
	for (retry = 0; retry < 2; retry++)
	{
	    int	h1, w1;
	    int	h2, w2;

	    if (font)
	    {
		DeleteObject(font);
		font = NULL;
	    }

	    if (font_bold)
	    {
		DeleteObject(font_bold);
		font_bold = NULL;
	    }

	    lf.lfWeight = 0;
	    font = CreateFontIndirect(&lf);

	    if (font != NULL && retry == 0)
	    {
		SelectObject(hdc, font);
		GetTextFace(hdc, sizeof(facename) - 1, facename);
		SelectObject(hdc, GetStockObject(SYSTEM_FONT));
		if (strcmp(lf.lfFaceName, DEFAULT_FONT) != 0 && strcmp(facename, lf.lfFaceName) != 0)
		{
		    retry = 0;
		    tty_fontname = DEFAULT_FONT;
		    strcpy(lf.lfFaceName, tty_fontname);
		    continue;
		}
	    }

	    lf.lfWeight = FW_BOLD;
	    font_bold = CreateFontIndirect(&lf);

	    SelectObject(hdc, font);
	    GetTextMetrics(hdc, &tm);
	    w1 = tm.tmAveCharWidth;
	    h1 = tm.tmHeight;

	    SelectObject(hdc, font_bold);
	    GetTextMetrics(hdc, &tm);
	    w2 = tm.tmAveCharWidth;
	    h2 = tm.tmHeight;

	    font_height = h1 > h2 ? h1 : h2;
	    font_width = w1 > w2 ? w1 : w2;

	    half_height = font_height / 2;
	    half_width = font_width / 2;

	    if (!mode_fix_window_size)
		break;

	    terminal_rect.left = 0;
	    terminal_rect.right = font_width * screen_width;
	    terminal_rect.top = 0;
	    terminal_rect.bottom = font_height * screen_height;

	    dx = ((client.right - client.left) - terminal_rect.right) / 2;
	    dy = ((client.bottom - client.top) - terminal_rect.bottom) / 2;

	    OffsetRect(&terminal_rect, dx, dy);

	    if (dx >= 0 && dy >= 0)
		break;

	    if (dx < 0)
		lf.lfWidth--;

	    if (dy < 0)
		lf.lfHeight--;
	}
    }

    for (dx = 0; dx < MAX_SCREEN_WIDTH; dx++)
	font_spacing[dx] = font_width;

    if (!mode_fix_window_size)
    {
	int		new_width, new_height;
	char		str[80];
	char		*p;

	new_width = (client.right - client.left) / font_width;
	new_height = (client.bottom - client.top) / font_height;

	new_width = max(new_width, 5);
	new_width = min(new_width, MAX_SCREEN_WIDTH);

	new_height = max(new_height, 4);
	new_height = min(new_height, MAX_SCREEN_HEIGHT);

	terminal_rect.left = 0;
	terminal_rect.right = font_width * new_width;
	terminal_rect.top = 0;
	terminal_rect.bottom = font_height * new_height;

	OffsetRect(&terminal_rect, SCREEN_BORDER, SCREEN_BORDER);

	if (new_height != screen_height || new_width != screen_width)
	{
	    tty_resize_screen(new_width, new_height);
	    p = send_tty_size(str);
	    port_write_str(str, p - str);
	}
    }

    /*
    ** Set default tab stops
    */
    tty_default_tab_stops();

    if (max_screen_history > 0)
    {
	SetScrollRange(hwndMain, SB_VERT, 0, max_screen_history, TRUE);
	tty_show_history(SB_BOTTOM, 0);
    }


    ReleaseDC(hWnd, hdc);
}

void	tty_clear_tab_stops(void)
{
    int		i;

    for (i = 0; i < MAX_SCREEN_WIDTH; i++)
    	tab_stops[i] = FALSE;
}

void	tty_default_tab_stops(void)
{
    int		i;

    for (i = 0; i < MAX_SCREEN_WIDTH; i++)
    {
	if (i > 0 && (i % 8) == 0)
	    tab_stops[i] = TRUE;
	else
	    tab_stops[i] = FALSE;
    }
}

void	tty_set_tab_stop(void)
{
    if (cursor_x > 0 && cursor_x < MAX_SCREEN_WIDTH)
	tab_stops[cursor_x] = TRUE;
}

void	tty_show_lastest(void)
{
    if (max_screen_history > 0 && screen_viewport_top < max_screen_history)
	tty_show_history(SB_BOTTOM, 0);
}

void	tty_show_history(int code, int pos)
{
    int		y;
    //char	buf[200];

    switch (code)
    {
    case SB_TOP:
    	pos = 0;
	break;

    case SB_PAGEUP:
    	pos = screen_viewport_top - screen_height;
	break;

    case SB_LINEUP:
    	pos = screen_viewport_top - 1;
	break;

    case SB_LINEDOWN:
    	pos = screen_viewport_top + 1;
	break;

    case SB_PAGEDOWN:
    	pos = screen_viewport_top + screen_height;
	break;

    case SB_BOTTOM:
    	pos = max_screen_history;
	break;

    case SB_THUMBPOSITION:
    case SB_THUMBTRACK:
	break;

    default:
	pos = screen_viewport_top;
	break;
    }

    if (pos < 0)
	pos = 0;

    if (pos > max_screen_history)
	pos = max_screen_history;

    screen_viewport_top = pos;
    screen_display = &screen_history[screen_viewport_top];
    for (y = 0; y < screen_height; y++)
    {
	t_line	*line = screen_display[y];
	line->modified = TRUE;
    }
    SetScrollPos(hwndMain, SB_VERT, screen_viewport_top, TRUE);
    tty_refresh(hwndMain, NULL);

    //sprintf(buf, "pos=%d", pos);
    //SetWindowText(hwndMain, buf);
}


void	tty_terminate(void)
{
    int		i;

    DeleteObject(pen_text);
    DeleteObject(pen_bold);
    DeleteObject(pen_cursor_text);
    DeleteObject(pen_background);
    DeleteObject(brush_text);
    DeleteObject(brush_bold);
    DeleteObject(brush_background);
    DeleteObject(brush_cursor);

    pen_text = NULL;
    pen_bold = NULL;
    pen_cursor_text = NULL;
    pen_background = NULL;
    brush_text = NULL;
    brush_bold = NULL;
    brush_background = NULL;
    brush_cursor = NULL;

    for (i = 0; i < MAX_TTY_COLOUR; i++)
    {
	DeleteObject(pen_colours[i]);
	DeleteObject(brush_colours[i]);
	pen_colours[i] = NULL;
	brush_colours[i] = NULL;
    }

    if (hScreenDC == NULL)
    {
	DeleteDC(hScreenDC);
	hScreenDC = NULL;
    }

    if (hScreenBitmap != NULL)
    {
	DeleteObject(hScreenBitmap);
	hScreenBitmap = NULL;
    }

}

static void do_move_cursor(int x, int y)
{
    x = max(x, 0);
    x = min(x, screen_width - 1);

    y = max(y, 0);
    y = min(y, screen_height - 1);

    cursor_hide(TRUE);

    cursor_x = x;
    cursor_y = y;

    vt_cursor_at_eol = FALSE;
    cursor_hide(FALSE);
}

static void do_shift_cursor(int diff)
{
    int		x = cursor_x;
    int 	y = cursor_y;

    x += diff;
    while (x < 0)
    {
	x += screen_width;
	y--;
    }

    while (x >= screen_width)
    {
	x -= screen_width;
	y++;
    }

    do_move_cursor(x, y);
}

static int	save_cursor_x = 0;
static int	save_cursor_y = 0;
static t_char	save_attribute = 0;

void	tty_save_state(void)
{
    save_cursor_x = cursor_x;
    save_cursor_y = cursor_y;
    save_attribute = curr_attribute;
}

void	tty_restore_state(void)
{
    curr_attribute = save_attribute;
    do_move_cursor(save_cursor_x, save_cursor_y);
}

void	tty_send_cursor_position(void)
{
    char	buf[128];

    sprintf(buf, "%c[%d;%dR", 0x1b, cursor_x + 1, cursor_y + 1);

    port_write_str(buf, strlen(buf));

}

void	tty_output_debug(char *fmt, ...)
{
   char		buf[256];
   char		*p;
   va_list 	argptr;

   va_start(argptr, fmt);
   vsprintf(buf, fmt, argptr);
   va_end(argptr);

   p = buf;

/*
   while (*p)
	tty_output(*p++);
*/
}

void	tty_output_str(char *fmt, ...)
{
   char		buf[256];
   char		*p;
   va_list 	argptr;

   va_start(argptr, fmt);
   vsprintf(buf, fmt, argptr);
   va_end(argptr);

   p = buf;

   while (*p)
	tty_output(*p++);
}

void	tty_output(int ch)
{
    tty_show_lastest();

    /*
    ** Handle special characters here
    */
    switch (ch)
    {
    case 0x09: /* Tab */
	{
	    int		x = cursor_x + 1;

	    while (x < screen_width)
	    {
		if (tab_stops[x])
		{
		    do_move_cursor(x, cursor_y);
		    return;
		}
		x++;
	    }
	}
	return;

    case 0x00: /* Ignore nulls */
	return;

    case 0x0c: /* Formfeed */
	do_move_cursor(0, 0);
	tty_clear_screen();
	return;

    case 0x08: /* Backspace */
	do_shift_cursor(-1);
	return;

    case 0x0a: /* Newline */
	tty_cursor_down_scroll();
	return;

    case 0x0d: /* Carriage Return */
	do_shift_cursor(-cursor_x);
	return;
    }

    /*
    ** Do not print control characters
    */
    if (ch < 32)
	return;

    if (vt_cursor_at_eol && (cursor_x == screen_width - 1))
    {
	if (cursor_y == screen_height - 1)
	{
	    tty_scroll_up();
	    do_shift_cursor(-cursor_x);
	}
	else
	{
	    do_shift_cursor(1);
	}
	vt_cursor_at_eol = FALSE;
    }

    if (cursor_x < screen_width && cursor_y < screen_height)
    {
	t_line	*line = screen_output[cursor_y];

	if (mode_insert)
	    tty_insert_chars(1);

	if (line->data[cursor_x] & ZATTR_BLINK)
	    line->blinks--;

	/*
	** If displaying over an existing selected text,
	** de-select all text first
	*/
	if (line->data[cursor_x] & ZATTR_SELECT)
	    tty_mouse_select(MOUSE_SELECT_START, -1, -1);

	line->data[cursor_x] = curr_attribute|(ch & (ZGR_MASK|0xff));
	line->modified = TRUE;

	if (curr_attribute & ZATTR_BLINK)
	    line->blinks++;

	if (cursor_x == screen_width - 1)
	{
	    vt_cursor_at_eol = TRUE;
	    cursor_hide(FALSE);
	}
	else
	{
	    vt_cursor_at_eol = FALSE;
	    do_shift_cursor(1);
	}
    }
}

void	tty_clear_screen(void)
{
    int		x, y;

    cursor_hide(TRUE);

    for (y = 0; y < screen_height; y++)
    {
	t_line	*line = screen_output[y];

	line->modified = TRUE;
	line->blinks = 0;
	for (x = 0; x < screen_width; x++)
	    line->data[x] = (curr_attribute & ZATTR_COLOUR_MASK) | ' ';
    }

    cursor_hide(FALSE);

    vt_cursor_at_eol = FALSE;
}

void	tty_clear_to_eol(void)
{
    int		x;
    t_line	*line = screen_output[cursor_y];

    cursor_hide(TRUE);

    line->modified = TRUE;
    for (x = cursor_x; x < screen_width; x++)
    {
	if (line->data[x] & ZATTR_BLINK)
	    line->blinks--;
	line->data[x] = (curr_attribute & ZATTR_COLOUR_MASK) | ' ';
    }

    cursor_hide(FALSE);
}

void	tty_clear_from_bol(void)
{
    int		x;
    t_line	*line = screen_output[cursor_y];

    cursor_hide(TRUE);

    line->modified = TRUE;
    for (x = cursor_x; x >= 0; x--)
    {
	if (line->data[x] & ZATTR_BLINK)
	    line->blinks--;
	line->data[x] = (curr_attribute & ZATTR_COLOUR_MASK) | ' ';
    }

    cursor_hide(FALSE);
}

void	tty_clear_to_eos(void)
{
    int		x;
    int		y;

    cursor_hide(TRUE);
    tty_clear_to_eol();

    for (y = cursor_y  + 1; y < screen_height; y++)
    {
	t_line	*line = screen_output[y];
	line->modified = TRUE;
	line->blinks = 0;
	for (x = 0; x < screen_width; x++)
	    line->data[x] = (curr_attribute & ZATTR_COLOUR_MASK) | ' ';
    }

    cursor_hide(FALSE);
}

void	tty_save_cursor(void)
{
    // FIXME
}

void	tty_restore_cursor(void)
{
    // FIXME
}

void	tty_cursor_home(void)
{
    do_move_cursor(0, 0);
}

void	tty_move_cursor_x(int x)
{
    do_move_cursor(x - 1, cursor_y);
}

void	tty_move_cursor(int x, int y)
{
    do_move_cursor(x - 1, y - 1);
}

void	tty_delete_chars(int n)
{
    t_line	*line = screen_output[cursor_y];
    int		x;

    if (n <= 0)
	n = 1;

    cursor_hide(TRUE);

    line->modified = TRUE;

    while (n-- > 0)
    {
	if (line->data[cursor_x] & ZATTR_BLINK)
	    line->blinks--;

	for (x = cursor_x; x < screen_width - 1; x++)
	    line->data[x] = line->data[x + 1];

	line->data[x] = (curr_attribute & ZATTR_COLOUR_MASK) | ' ';
    }

    cursor_hide(FALSE);
}

void	tty_set_insert_mode(int insert)
{
    mode_insert = insert;
}

static void	set_colours()
{
    curr_attribute &= ~ZATTR_COLOUR_MASK;

    if (curr_foreground_colour >= 0)
    {
	curr_attribute |= ZATTR_COLOUR_FG_SET;
	curr_attribute |=(curr_foreground_colour << 24);
    }

    if (curr_background_colour >= 0)
    {
	curr_attribute |= ZATTR_COLOUR_BG_SET;
	curr_attribute |= (curr_background_colour << 28);
    }
}

void	tty_foreground_colour(int colour)
{
    if (colour > 0)
	colour &= 0x07;

    curr_foreground_colour = colour;

    set_colours();
}

void	tty_background_colour(int colour)
{
    if (colour > 0)
	colour &= 0x07;

    curr_background_colour = colour;

    set_colours();
}

void	tty_insert_chars(int n)
{
    t_line	*line = screen_output[cursor_y];
    int		x;

    if (n <= 0)
	n = 1;

    cursor_hide(TRUE);

    line->modified = TRUE;

    while (n-- > 0)
    {
	if (line->data[screen_width - 1] & ZATTR_BLINK)
	    line->blinks--;

	for (x = screen_width - 1; x > cursor_x; x--)
	    line->data[x] = line->data[x - 1];

	line->data[cursor_x] = (curr_attribute & ZATTR_COLOUR_MASK) | ' ';
    }

    cursor_hide(FALSE);
}

void	tty_delete_lines(int n)
{
    int		y;

    cursor_hide(TRUE);

    while (n-- > 0 && cursor_y < scroll_bottom)
    {
	t_line	*l1 = screen_output[cursor_y];

	for (y = cursor_y; y < scroll_bottom; y++)
	{
	    screen_output[y] = screen_output[y + 1];
	    screen_output[y]->modified = TRUE;
	}

	screen_output[y] = l1;

	clear_line(l1);
    }

    cursor_hide(FALSE);
}

void	tty_insert_lines(int n)
{
    int		y;

    cursor_hide(TRUE);

    while (n-- > 0 && cursor_y < scroll_bottom)
    {
	t_line	*l1 = screen_output[scroll_bottom];

	for (y = scroll_bottom; y > cursor_y; y--)
	{
	    screen_output[y] = screen_output[y - 1];
	    screen_output[y]->modified = TRUE;
	}

	screen_output[y] = l1;

	clear_line(l1);
    }

    cursor_hide(FALSE);
}

void	tty_cursor_up_scroll(void)
{
    if (cursor_y > scroll_top)
	do_shift_cursor(-screen_width);
    else
	tty_scroll_down();
}

void	tty_cursor_down_scroll(void)
{
    if (cursor_y < scroll_bottom)
	do_shift_cursor(screen_width);
    else
	tty_scroll_up();
}


void	tty_cursor_up(int x)
{
    if (x <= 0)
	x = 1;
    do_shift_cursor(-(screen_width * x));
}

void	tty_cursor_down(int x)
{
    if (x <= 0)
	x = 1;
    do_shift_cursor(screen_width * x);
}

void	tty_cursor_left(int x)
{
    if (x <= 0)
	x = 1;
    do_shift_cursor(-x);
}

void	tty_cursor_right(int x)
{
    if (x <= 0)
	x = 1;
    do_shift_cursor(x);
}


int	tty_get_attr(void)
{
    return curr_attribute;
}

void	tty_set_attr(int attr)
{
    curr_attribute = (curr_attribute & ZATTR_COLOUR_MASK) | attr;
}


static void	tty_refreshX(HWND hWnd, HDC hdc, BOOL do_blink)
{
    int		x, xend, y;
    int		sx, sy;
    BOOL	must_release_dc = FALSE;
    int		dc_setup = FALSE;
    COLORREF	curr_colour = 0;
    COLORREF	curr_background = 0;
    HBRUSH	curr_brush = NULL;
    HPEN	curr_pen = NULL;
    HFONT	curr_font = NULL;


    sy = terminal_rect.top;
    for (y = 0; y < screen_height; y++, sy += font_height)
    {
	t_line	*line = screen_display[y];
	int	update_x0 = 0;
	int	update_x1 = 0;

	if (!line->modified && (!do_blink || line->blinks == 0))
	    continue;

	line->modified = FALSE;

	sx = terminal_rect.left;

	for (x = 0; x < screen_width; sx += (xend - x) * font_width, x = xend)
	{
	    int		cattr;
	    int		battr;
	    int		gattr;
	    t_char	item = line->data[x];
	    RECT	crect;
	    HBRUSH	zbrush;
	    HPEN	zpen;
	    COLORREF	zcolour;
	    COLORREF	zbackground;
	    HFONT	zfont;

	    xend = x + 1;


	    if (screen_contents[x][y] == item && (!do_blink || !(item & ZATTR_BLINK)))
		continue;

	    if (hdc == NULL)
	    {
		must_release_dc = TRUE;
		hdc = GetDC(hWnd);
		set_palette(hdc);
	    }

	    cattr = item & ZATTR_MASK;
	    gattr = item & ZGR_MASK;

	    if (cattr & ZATTR_SELECT)
	    {
		cattr &= ~ZATTR_SELECT;
		cattr ^= ZATTR_REVERSE;
	    }

	    if (!blink_shown && (cattr & ZATTR_BLINK) && 
	    	(cattr & ZATTR_BOLD) && ((cattr & ZATTR_REVERSE) == 0))
		battr = cattr ^ ZATTR_BOLD;
	    else
		battr = cattr;

	    if (battr & ZATTR_CURSOR)
	    {
		zbrush = brush_cursor;
		zpen = pen_cursor_text;
		zcolour = colour_cursor_text;
		zbackground = colour_cursor;
		zfont = font;
	    }
	    else if (battr & ZATTR_REVERSE)
	    {
		if (battr & (ZATTR_COLOUR_FG_SET|ZATTR_COLOUR_BG_SET))
		{
		    zbrush = brush_text;
		    zpen = pen_background;
		    zcolour = colour_background;
		    zbackground = colour_text;

		    if (battr & ZATTR_COLOUR_FG_SET)
		    {
			int	bg = ZATTR_COLOUR_FG(battr);

			zbrush = brush_colours[bg];
			zbackground = tty_colours[bg];
		    }

		    if (battr & ZATTR_COLOUR_BG_SET)
		    {
			int	fg = ZATTR_COLOUR_BG(battr);

			if (battr & ZATTR_BOLD)
			    fg += 8;

			zpen = pen_colours[fg];
			zcolour = tty_colours[fg];
		    }

		    zfont = (battr & ZATTR_BOLD) ? font_bold : font;
		}
		else if (battr & ZATTR_BOLD)
		{
		    zbrush = brush_bold;
		    zpen = pen_background;
		    zcolour = colour_background;
		    zbackground = colour_bold;
		    zfont = font_bold;
		}
		else
		{
		    zbrush = brush_text;
		    zpen = pen_background;
		    zcolour = colour_background;
		    zbackground = colour_text;
		    zfont = font;
		}
	    }
	    else
	    {
		if (battr & (ZATTR_COLOUR_FG_SET|ZATTR_COLOUR_BG_SET))
		{
		    zbrush = brush_background;
		    zpen = pen_text;
		    zcolour = colour_text;
		    zbackground = colour_background;

		    if (battr & ZATTR_COLOUR_BG_SET)
		    {
			int	bg = ZATTR_COLOUR_BG(battr);

			zbrush = brush_colours[bg];
			zbackground = tty_colours[bg];
		    }

		    if (battr & ZATTR_COLOUR_FG_SET)
		    {
			int	fg = ZATTR_COLOUR_FG(battr);

			if (battr & ZATTR_BOLD)
			    fg += 8;

			zpen = pen_colours[fg];
			zcolour = tty_colours[fg];
		    }

		    zfont = (battr & ZATTR_BOLD) ? font_bold : font;
		}
		else if (battr & ZATTR_BOLD)
		{
		    zbrush = brush_background;
		    zpen = pen_bold;
		    zcolour = colour_bold;
		    zbackground = colour_background;
		    zfont = font_bold;
		}
		else
		{
		    zbrush = brush_background;
		    zpen = pen_text;
		    zcolour = colour_text;
		    zbackground = colour_background;
		    zfont = font;
		}
	    }


	    if (!dc_setup)
	    {
		curr_font = zfont;
		curr_colour = zcolour;
		curr_background = zbackground;

		SetBkMode(hScreenDC, OPAQUE);
		SetBkColor(hScreenDC, zbackground);
		SelectObject(hScreenDC, curr_font);
		SetTextAlign(hScreenDC, TA_TOP|TA_LEFT);
		SetTextColor(hScreenDC, zcolour);
		dc_setup = TRUE;
	    }

	    screen_contents[x][y] = item;

	    if (!gattr)
	    {
		while (xend < screen_width)
		{
		    t_char	c = line->data[xend];

		    if (c == screen_contents[xend][y])
			break;

		    if (c & ZGR_MASK)
			break;

		    if ((c & ZATTR_MASK) != cattr)
			break;

		    screen_contents[xend][y] = c;
		    xend++;
		}
	    }

	    crect.left = sx;
	    crect.right = sx + ((xend - x) * font_width);
	    crect.top = sy;
	    crect.bottom = sy + font_height;

	    if (!blink_shown && (cattr & ZATTR_BLINK) && (cattr == battr))
	    {
		FillRect(hScreenDC, &crect, zbrush);
		FillRect(hdc, &crect, zbrush);
		continue;
	    }

	    if (gattr)
	    {
		if (curr_pen != zpen)
		{
		    SelectObject(hScreenDC, zpen);
		    curr_pen = zpen;
		}

		FillRect(hScreenDC, &crect, zbrush);

		switch (gattr)
		{
		case ZGR_H_LINE:
		    MoveToEx(hScreenDC, sx, sy + half_height, NULL);
		    LineTo(hScreenDC, sx + font_width, sy + half_height);
		    break;

		case ZGR_V_LINE:
		    MoveToEx(hScreenDC, sx + half_width, sy, NULL);
		    LineTo(hScreenDC, sx +  + half_width, sy + font_height);
		    break;

		case ZGR_TL_CORNER:
		    MoveToEx(hScreenDC, sx + font_width, sy + half_height, NULL);
		    LineTo(hScreenDC, sx + half_width, sy + half_height);
		    LineTo(hScreenDC, sx + half_width, sy + font_height);
		    break;

		case ZGR_TR_CORNER:
		    MoveToEx(hScreenDC, sx, sy + half_height, NULL);
		    LineTo(hScreenDC, sx + half_width, sy + half_height);
		    LineTo(hScreenDC, sx + half_width, sy + font_height);
		    break;

		case ZGR_BL_CORNER:
		    MoveToEx(hScreenDC, sx + half_width, sy, NULL);
		    LineTo(hScreenDC, sx + half_width, sy + half_height);
		    LineTo(hScreenDC, sx + font_width, sy + half_height);
		    break;

		case ZGR_BR_CORNER:
		    MoveToEx(hScreenDC, sx + half_width, sy, NULL);
		    LineTo(hScreenDC, sx + half_width, sy + half_height);
		    LineTo(hScreenDC, sx - 1, sy + half_height);
		    break;

		case ZGR_TEE_LEFT:
		    MoveToEx(hScreenDC, sx + half_width, sy, NULL);
		    LineTo(hScreenDC, sx +  + half_width, sy + font_height);
		    MoveToEx(hScreenDC, sx, sy + half_height, NULL);
		    LineTo(hScreenDC, sx + half_width, sy + half_height);
		    break;

		case ZGR_TEE_RIGHT:
		    MoveToEx(hScreenDC, sx + half_width, sy, NULL);
		    LineTo(hScreenDC, sx +  + half_width, sy + font_height);
		    MoveToEx(hScreenDC, sx + half_width, sy + half_height, NULL);
		    LineTo(hScreenDC, sx + font_width, sy + half_height);
		    break;

		case ZGR_TEE_TOP:
		    MoveToEx(hScreenDC, sx, sy + half_height, NULL);
		    LineTo(hScreenDC, sx + font_width, sy + half_height);
		    MoveToEx(hScreenDC, sx + half_width, sy, NULL);
		    LineTo(hScreenDC, sx +  + half_width, sy + half_height);
		    break;

		case ZGR_TEE_BOTTOM:
		    MoveToEx(hScreenDC, sx, sy + half_height, NULL);
		    LineTo(hScreenDC, sx + font_width, sy + half_height);
		    MoveToEx(hScreenDC, sx + half_width, sy + half_height, NULL);
		    LineTo(hScreenDC, sx +  + half_width, sy + font_height);
		    break;

		case ZGR_CROSS:
		    MoveToEx(hScreenDC, sx, sy + half_height, NULL);
		    LineTo(hScreenDC, sx + font_width, sy + half_height);
		    MoveToEx(hScreenDC, sx + half_width, sy, NULL);
		    LineTo(hScreenDC, sx +  + half_width, sy + font_height);
		    break;
		}
	    }
	    else
	    {
		unsigned char	str[MAX_SCREEN_WIDTH];
		int		i;
		t_char		ch;

		for (i = x; i < xend; i++)
		{
		    ch = line->data[i];
		    str[i - x] = ch & 0xff;
		}

		if (curr_colour != zcolour)
		{
		    SetTextColor(hScreenDC, zcolour);
		    curr_colour = zcolour;
		}

		if (curr_background != zbackground)
		{
		    SetBkColor(hScreenDC, zbackground);
		    curr_background = zbackground;
		}


		if (curr_font != zfont)
		{
		    curr_font = zfont;
		    SelectObject(hScreenDC, curr_font);
		}

		ExtTextOut(hScreenDC, sx, sy, ETO_OPAQUE|ETO_CLIPPED, &crect, str, xend - x, font_spacing);

		if (cattr & ZATTR_UNDERLINE)
		{
		    if (curr_pen != zpen)
		    {
			SelectObject(hScreenDC, zpen);
			curr_pen = zpen;
		    }
		    MoveToEx(hScreenDC, crect.left, crect.bottom - 1, NULL);
		    LineTo(hScreenDC, crect.right, crect.bottom - 1);
		}
	    }

	    BitBlt(hdc, sx, sy, (xend - x) * font_width, font_height,
	    	hScreenDC, sx, sy, SRCCOPY);
	}
    }

    if (must_release_dc)
	ReleaseDC(hWnd, hdc);

}

static BOOL	do_blink_requested = FALSE;
static BOOL	in_refresh = FALSE;

void	tty_refresh(HWND hWnd, HDC hdc)
{
    if (!in_refresh)
    {
	in_refresh = TRUE;

	if (hdc)
	    set_palette(hdc);

	tty_refreshX(hWnd, hdc, do_blink_requested);

	do_blink_requested = FALSE;

	in_refresh = FALSE;
    }
}

void	tty_repaint(HWND hWnd, HDC hdc)
{
    BOOL	must_release_dc = FALSE;
    RECT	client;

    if (!screen_display)
	return;

    if (!font)
	return;

    if (hdc == NULL)
    {
	must_release_dc = TRUE;
	hdc = GetDC(hWnd);
    }

    set_palette(hdc);
    GetClientRect(hWnd, &client);

    BitBlt(hdc, 
    	client.left, client.top, client.right - client.left, client.bottom - client.top,
	hScreenDC, 0, 0, SRCCOPY);

    if (must_release_dc)
	ReleaseDC(hWnd, hdc);
}

void	tty_do_blink(void)
{
    blink_shown = !blink_shown;

    if (in_refresh)
    {
	do_blink_requested = TRUE;
	return;
    }

    if (hwndMain)
	tty_refreshX(hwndMain, NULL, TRUE);

    do_blink_requested = FALSE;
}

static void	mouse_to_pos(int *x, int *y)
{
    *x = (*x - terminal_rect.left) / font_width;
    *x = max(0, *x);
    *x = min(screen_width, *x);

    *y = (*y - terminal_rect.top) / font_height;
    *y = max(0, *y);
    *y = min(screen_height, *y);

    if (*x < screen_width && *y < screen_height)
    {
	t_line	*line = screen_display[*y];
	int	all_spaces = TRUE;
	int	i;

	for (i = *x; i < screen_width; i++)
	{
	    if ((line->data[i] & ~(ZATTR_SELECT| ZATTR_BOLD | ZATTR_CURSOR)) != ' ')
	    {
		all_spaces = FALSE;
		break;
	    }
	}

	if (all_spaces)
	{
	    *x = 0;
	    *y = *y + 1;

	}
    }
}

static int	select_start_x = 0;
static int	select_start_y = 0;

static int	select_end_x = -1;
static int	select_end_y = -1;

static int	select_extend_x = -1;
static int	select_extend_y = -1;

static void	get_selected_range(int *start, int *end)
{
    int		s0 = select_start_x + (select_start_y * screen_width);
    int		s1 = select_end_x + (select_end_y * screen_width);
    int		s2 = select_extend_x + (select_extend_y * screen_width);

    *end = max(s0, s1);
    *end = min(*end, screen_height * screen_width);

    *start = min(s0, s1);

    if (s2 >= 0)
    {
	*start = min(*start, s2);
	*end = max(*end, s2);
    }

}

static void	copy_selected_text(void)
{
    int		z0, z1, z;
    int		x, y;
    t_char	ch;

    get_selected_range(&z0, &z1);

    if (z0 >= 0 && z1 > z0)
    {
	int	size = (screen_width + 2) * (abs((z1 / screen_width) - (z0 / screen_width)) + 1) + 1;
	HGLOBAL	handle = GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE, size);
	char	*s = GlobalLock(handle);
	char	*p = s;

	for (z = z0; z < z1; z++)
	{
	    x = z % screen_width;
	    y = z / screen_width;
	    ch = screen_display[y]->data[x];

	    if (ch & ZGR_MASK)
		*p++ = ' ';
	    else
		*p++ = ch & 0xdff;

	    if (z > z0 && x == screen_width - 1)
	    {
		if (p > s)
		{
		    p--;
		    while (p >= s && *p == ' ')
			p--;
		    p++;
		};
		*p++ = 0x0d;
		*p++ = 0x0a;
	    }
	}

	*p = '\0';

	GlobalUnlock(handle);
	if (OpenClipboard(hwndMain))
	{
	    EmptyClipboard();
	    SetClipboardData(CF_OEMTEXT, handle);
	    CloseClipboard();
	}
	else
	{
	    GlobalFree(handle);
	}

	select_start_x = z0 % screen_width;
	select_start_y = z0 / screen_width;
	select_end_x = z1 % screen_width;
	select_end_y = z1 / screen_width;
	select_extend_x = -1;
	select_extend_y = -1;
    }
}

static void	select_text(void)
{
    int		z0, z1;
    int		x, y;
    int		z;

    get_selected_range(&z0, &z1);

    for (y = 0; y < screen_height; y++)
    {
	t_line	*line = screen_display[y];

	z = y * screen_width;

	for (x = 0; x < screen_width; x++, z++)
	{
	    t_char	ch = line->data[x];

	    if (z >= z0 && z < z1)
	    {
		if (!(ch & ZATTR_SELECT))
		{
		    line->modified = TRUE;
		    line->data[x] |= ZATTR_SELECT;
		}
	    }
	    else
	    {
		if (ch & ZATTR_SELECT)
		{
		    line->modified = TRUE;
		    line->data[x] &= ~ZATTR_SELECT;
		}
	    }
	}
    }

    tty_refresh(hwndMain, NULL);

}

void tty_mouse_select(int mode, int x, int y)
{
    mouse_to_pos(&x, &y);
    switch (mode)
    {
    case MOUSE_SELECT_START:
	select_start_x = x;
	select_start_y = y;
	select_end_x = x;
	select_end_y = y;
	select_extend_x = -1;
	select_extend_y = -1;
	select_text();
	break;

    case MOUSE_SELECT_MOVE:
	select_end_x = x;
	select_end_y = y;
	select_text();
	break;

    case MOUSE_SELECT_END:
	select_text();
	copy_selected_text();
	break;

    case MOUSE_SELECT_EXTENT:
	select_extend_x = x;
	select_extend_y = y;
	select_text();
	break;

    case MOUSE_SELECT_WORD:
	if (x < screen_width && y < screen_height)
	{
	    t_line	*line = screen_display[y];

	    select_extend_x = -1;
	    select_extend_y = -1;

	    select_start_x = x;
	    select_start_y = y;

	    select_end_x = x;
	    select_end_y = y;

	    if ((line->data[select_start_x] & ~(ZATTR_SELECT| ZATTR_BOLD | ZATTR_CURSOR)) == ' ')
		return;

	    while (select_start_x > 0)
	    {
		if ((line->data[select_start_x - 1] & ~(ZATTR_SELECT| ZATTR_BOLD | ZATTR_CURSOR)) == ' ')
		    break;
		select_start_x--;
	    }

	    while (select_end_x < screen_width)
	    {
		if ((line->data[select_end_x] & ~(ZATTR_SELECT| ZATTR_BOLD | ZATTR_CURSOR)) == ' ')
		    break;
		select_end_x++;
	    }

	    if (select_start_x < select_end_x)
	    {
		select_text();
		copy_selected_text();
	    }
	}
	break;
    }
}

/**
 * Get the DISPLAY environment variable.
 *
 * If the variable is of the form "localhost:..",
 * replace 'localhost' with the actual ip-address.
 *
 */
static char	*get_xdisplay(void)
{
    static char 	xdisp[255];
    char		hostname[128];
    char		*display;
    struct hostent	*hostinfo;

    if (xdisp[0] != '\0')
	return xdisp;

    display = getenv("DISPLAY");

    if (display == NULL)
	return NULL;

    if (strlen(display) > 10 &&
    	strncmp(display, "localhost:", 10) == 0 &&
	gethostname(hostname, sizeof(hostname) - 1) == 0 &&
	(hostinfo = gethostbyname(hostname)) != NULL)
    {
	sprintf(xdisp, "%d.%d.%d.%d%s",
	    (unsigned char)hostinfo->h_addr[0], (unsigned char)hostinfo->h_addr[1],
	    (unsigned char)hostinfo->h_addr[2], (unsigned char)hostinfo->h_addr[3],
	    display + 9);

	//strncpy(xdisp, hostname, sizeof(xdisp) - strlen(display) - 12);
	//strcat(xdisp, display + 9);
    }
    else
    {
	strncpy(xdisp, display, sizeof(xdisp) - 1);
    }

    return xdisp;
}

int tty_telnet_controls(unsigned char *buf, int size)
{
    int		i = 0;
    int		j;
    char	respond[80];
    char	*p_respond = respond;

    while (i < size - 2)
    {
	if (buf[i] == IAC)
	{
	    int		skip = 0;

	    switch (buf[i + 1])
	    {
	    case SB:
		tty_output_debug("<SB %d %d>\n\r", buf[i + 2], buf[i + 3]);
		for (j = i + 2; j < size - 1; j++)
		{
		    if (buf[j] == IAC && buf[j + 1] == SE)
		    {
			skip = j - i + 2;
			break;
		    }
		}

		if (skip > 0 && buf[i + 3] == TELQUAL_SEND)
		{
		    char	*tt;

		    switch (buf[i + 2])
		    {
		    case TELOPT_TTYPE:
			tty_output_debug("<send terminal type>");
			*p_respond++ = IAC;
			*p_respond++ = SB;
			*p_respond++ = TELOPT_TTYPE;
			*p_respond++ = TELQUAL_IS;
			tt = terminal_type; 
			while (*tt != '\0')
			    *p_respond++ = *tt++;
			*p_respond++ = IAC;
			*p_respond++ = SE;
			break;

		    case TELOPT_XDISPLOC:
			{
			    char	*display = get_xdisplay();

			    if (display == NULL)
				display = "none";

			    tty_output_debug("<send x-display>");
			    *p_respond++ = IAC;
			    *p_respond++ = SB;
			    *p_respond++ = TELOPT_XDISPLOC;
			    *p_respond++ = TELQUAL_IS;
			    while (*display)
				*p_respond++ = *display++;
			    *p_respond++ = IAC;
			    *p_respond++ = SE;
			}
			break;
		    }
		}
		break;

	    case DO:
		tty_output_debug("<DO %d>", buf[i + 2]);
		skip = 3;
		switch (buf[i + 2])
		{
		case TELOPT_TTYPE:
		    *p_respond++ = IAC;
		    *p_respond++ = WILL;
		    *p_respond++ = buf[i + 2];
		    break;

		case TELOPT_NAWS:
		    *p_respond++ = IAC;
		    *p_respond++ = WILL;
		    *p_respond++ = buf[i + 2];
		    p_respond = send_tty_size(p_respond);
		    break;

		case TELOPT_XDISPLOC:
		    *p_respond++ = IAC;
		    *p_respond++ = get_xdisplay() ? WILL : WONT;
		    *p_respond++ = buf[i + 2];
		    break;

		case TELOPT_ECHO:
		    *p_respond++ = IAC;
		    *p_respond++ = WONT;
		    *p_respond++ = buf[i + 2];
		    break;

		default:
		    *p_respond++ = IAC;
		    *p_respond++ = WONT;
		    *p_respond++ = buf[i + 2];
		    break;
		}
		break;

	    case DONT:
		tty_output_debug("<DONT %d>", buf[i + 2]);
		skip = 3;
		switch (buf[i + 1])
		{
		case TELOPT_ECHO:
		    mode_echo = FALSE;
		    break;
		}
		break;

	    case WILL:
		tty_output_debug("<WONT %d>", buf[i + 2]);
		skip = 3;
		break;

	    case WONT:
		tty_output_debug("<WONT %d>", buf[i + 2]);
		skip = 3;
		break;

	    default:
		//_tprintf("Unknown IAC %x %x\n", buf[i + 1], buf[i + 2]);
		break;
	    }

	    if (skip > 0)
	    {
		memmove(buf + i, buf + i + skip, size - i - skip);
		size -= skip;
		continue;
	    }
	}
	i++;
    }

    if (p_respond > respond)
	port_write_str(respond, p_respond - respond);

    return size;
}

