
#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <winsock2.h>
#include "vt.h"
#include "emulator.h"

HINSTANCE	hInst;
HWND		hwndMain;
SOCKET		telnet_socket;
WSADATA		WsaData;

#define VT_TERM_WND_CLASS	"VtTermWndClass"

LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

static BOOL	allow_close = TRUE;

static int	telnet_port = 23;
static char	*telnet_host = NULL;

static int	blink_step = 0;
static int	term_need_refresh = FALSE;

#define	WINDOW_NAME	"VT"

/*
** Display error message
*/
void 	error_message(char *fmt, ...)
{
   char		buf[256];
   va_list 	argptr;

   va_start(argptr, fmt);
   vsprintf(buf, fmt, argptr);
   va_end(argptr);
   MessageBox(hwndMain, buf, "ERROR", MB_ICONSTOP);
}

/*
** Display error message and exit
*/
void 	error_exit(char *fmt, ...)
{
   char		buf[256];
   va_list 	argptr;

   va_start(argptr, fmt);
   vsprintf(buf, fmt, argptr);
   va_end(argptr);
   error_message(buf);
   allow_close = TRUE;
   PostMessage(hwndMain,WM_CLOSE,0,0);
}

/*
** Write a single character to the socket
*/
void port_write(char ch)
{
    if (telnet_socket)
	send(telnet_socket, &ch, 1, 0);
}

/*
** Write a string to the socket
*/
void port_write_str(unsigned char *s, int len)
{
    if (telnet_socket)
	send(telnet_socket, s, len, 0);
}

static BOOL InitApplication(void)
{
	WNDCLASS wc;


	memset(&wc,0,sizeof(WNDCLASS));
	wc.style = CS_DBLCLKS | CS_OWNDC;
	wc.lpfnWndProc = (WNDPROC)MainWndProc;
	wc.hInstance = hInst;
	wc.hbrBackground = NULL;
	wc.lpszClassName = VT_TERM_WND_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL,IDC_IBEAM);
	wc.hIcon = LoadIcon(NULL,IDI_APPLICATION);

	if (!RegisterClass(&wc))
		return 0;

	return 1;
}

HWND CreateVtTermWndClassWnd(void)
{
    int		flags =  WS_MINIMIZEBOX | WS_VISIBLE | WS_MAXIMIZEBOX | WS_CAPTION |
    			 WS_BORDER | WS_SYSMENU | WS_THICKFRAME;

    return CreateWindow(VT_TERM_WND_CLASS, WINDOW_NAME, flags,
		CW_USEDEFAULT,0,CW_USEDEFAULT,0,
		NULL,
		NULL,
		hInst,
		NULL);
}

#define IDTIMER	123

static BOOL	k_shift_down = FALSE;
static BOOL	k_control_down = FALSE;

static int key_used(int key)
{
    switch (key)
    {
    case VK_SHIFT:
    case VK_CONTROL:
    case VK_F1:
    case VK_F2:
    case VK_F3:
    case VK_F4:
    case VK_F5:
    case VK_F6:
    case VK_F7:
    case VK_F8:
    case VK_F9:
    case VK_F10:
    case VK_F11:
    case VK_F12:
    case VK_LEFT:
    case VK_RIGHT:
    case VK_UP:
    case VK_DOWN:
    case VK_PRIOR:
    case VK_NEXT:
    case VK_END:
    case VK_HOME:
    case VK_DELETE:
    case VK_INSERT:
    case VK_PAUSE:
    case VK_BACK:
	return TRUE;

    default:
	return FALSE;
    }
}

static BOOL do_fkey(int key)
{
    int		fkey = 0;

    if (k_control_down && !k_shift_down)
    {
	switch (key)
	{
	case VK_F1:	fkey = ZKEY_F13;	break;
	case VK_F2:	fkey = ZKEY_F14;	break;
	case VK_F9:	fkey = ZKEY_F17;	break;
	case VK_F10:	fkey = ZKEY_F18;	break;
	}
    }
    else if (!k_control_down && !k_shift_down)
    {
	switch (key)
	{
	case VK_F1:	fkey = ZKEY_PF1;	break;
	case VK_F2:	fkey = ZKEY_PF2;	break;
	case VK_F3:	fkey = ZKEY_PF3;	break;
	case VK_F4:	fkey = ZKEY_PF4;	break;

	case VK_F5:	fkey = ZKEY_F16; break;

	case VK_F6:	fkey = ZKEY_F6;	break;
	case VK_F7:	fkey = ZKEY_F7;	break;
	case VK_F8:	fkey = ZKEY_F8;	break;
	case VK_F9:	fkey = ZKEY_F9;	break;
	case VK_F10:	fkey = ZKEY_F10; break;
	case VK_F11:	fkey = ZKEY_F11; break;
	case VK_F12:	fkey = ZKEY_F12; break;

	case VK_LEFT:	fkey = ZKEY_LEFT; break;
	case VK_RIGHT:	fkey = ZKEY_RIGHT; break;
	case VK_UP:	fkey = ZKEY_UP; break;
	case VK_DOWN:	fkey = ZKEY_DOWN; break;
	case VK_PRIOR:	fkey = ZKEY_PGUP; break;
	case VK_NEXT:	fkey = ZKEY_PGDOWN; break;
	case VK_END:	fkey = ZKEY_SELECT; break;
	case VK_HOME:	fkey = ZKEY_FIND; break;
	case VK_DELETE:	fkey = ZKEY_FIND; break;
	case VK_INSERT:	fkey = ZKEY_INSERT; break;
	case VK_PAUSE:	fkey = ZKEY_F20; break;
	case VK_BACK:	fkey = ZKEY_BACKSPACE; break;
	}
    }
    else if (!k_control_down && k_shift_down)
    {
	switch (key)
	{
	case VK_F1:
	case VK_F2:
	case VK_F3:
	case VK_F4:
	case VK_F5:
	case VK_F6:
	case VK_F7:
	case VK_F8:
	case VK_F9:
	case VK_F10:
	case VK_F11:
	case VK_F12:
	    break;
	}
    }

    if (!fkey)
	return FALSE;

    key_translate(fkey);
    return TRUE;
}

static BOOL	in_select = FALSE;
static BOOL	in_extent_select = FALSE;
static BOOL	has_3_button_mouse = FALSE;
static BOOL	allow_select = TRUE;
static BOOL	paste_newline = TRUE;

static void do_paste(HWND hWnd)
{
    if (OpenClipboard(hWnd))
    {
	HANDLE	handle;
	char	*text;
	int	len;

	handle = GetClipboardData(CF_TEXT);
	if (handle)
	{
	    text = GlobalLock(handle);
	    len = strlen(text);

	    if (paste_newline)
	    {
		port_write_str(text, len);
	    }
	    else
	    {
		int	i;
		int	ch;
		int	was_space = FALSE;
		char	*buf, *q;

		buf = malloc(len + 1);
		q = buf;
		for (i = 0; i < len; i++)
		{
		    ch = text[i];
		    if (isspace(ch))
		    {
			if (!was_space)
			{
			    *q++ = ' ';
			    was_space = TRUE;
			}
			continue;
		    }
		    else
		    {
			was_space = FALSE;
		    }
		    *q++ = ch;
		}
		*q = '\0';
		port_write_str(buf, strlen(buf));
		free(buf);
	    }

	    GlobalUnlock(handle);
	}
	CloseClipboard();

    }
}

LRESULT CALLBACK MainWndProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
    case WM_CREATE:
	has_3_button_mouse = (GetSystemMetrics(SM_CMOUSEBUTTONS) >= 3);
	break;

    case WM_CLOSE:
	if (allow_close)
	    DestroyWindow(hWnd);
	return 0;

    case WM_DESTROY:
    	KillTimer(hWnd, IDTIMER);
	PostQuitMessage(0);
	break;

    case WM_KEYUP:
	switch (wParam)
	{
	case VK_SHIFT:
	    k_shift_down = FALSE;
	    break;

	case VK_CONTROL:
	    k_control_down = FALSE;
	    break;
	}
	break;

    case WM_SYSCOMMAND:
	switch (wParam)
	{
	case SC_KEYMENU:
	    do_fkey(VK_F10);
	    return 0;
	}
	break;

    case WM_KEYDOWN:
	switch (wParam)
	{
	case VK_SHIFT:
	    k_shift_down = TRUE;
	    break;

	case VK_CONTROL:
	    k_control_down = TRUE;
	    break;

	default:
	    if (do_fkey(wParam))
		return 0;
	    break;
	}
	break;

    case WM_CHAR:
	{
	    int		key = wParam;

	    if (!key_translate(key))
	    {
		port_write(key);
		tty_echo(key);
	    }
	}
	break;

    case WM_TIMER:
	blink_step++;
	if (blink_step >= 5)
	{
	    blink_step = 0;
	    tty_do_blink();
	    term_need_refresh = FALSE;
	}
	if (term_need_refresh)
	{
	    term_need_refresh = FALSE;
	    tty_refresh(hwndMain, NULL);
	}
	break;

    case WM_SIZE:
	switch (wParam)
	{
	case SIZE_MAXIMIZED:
	case SIZE_RESTORED:
	    tty_init(hWnd);
	    tty_repaint(hWnd, NULL);
	    return 0;
	}
	break;

    case WM_PAINT:
	{
	    PAINTSTRUCT	ps;

	    BeginPaint(hWnd, &ps);
	    tty_repaint(hWnd, ps.hdc);
	    EndPaint(hWnd, &ps);
	}
	return 0;

    case WM_LBUTTONDOWN:
	if (allow_select && !in_select && !in_extent_select)
	{
	    in_select = TRUE;
	    SetCapture(hWnd);
	    tty_mouse_select(MOUSE_SELECT_START, LOWORD(lParam), HIWORD(lParam));
	}
	return 0;

#define WHEEL_DELTA 120

    case WM_MOUSEWHEEL:
	if (max_screen_history > 0)
	{
	    int		delta = HIWORD(wParam) / WHEEL_DELTA;
	    int		pos = GetScrollPos(hWnd, SB_VERT);

	    if (delta != 0)
		tty_show_history(SB_THUMBPOSITION, pos - delta);
	}


    case WM_VSCROLL:
	if (max_screen_history > 0)
	{
	    tty_show_history(LOWORD(wParam), HIWORD(wParam));
	}
    	return 0;

    case WM_RBUTTONDOWN:
	if (allow_select && has_3_button_mouse)
	{
	    /*
	    ** For 3 button mouse, the right mouse button is use for extending the selection
	    ** For 2 button mouse, the right mouse button is use for pasting
	    */
	    if (!in_select && !in_extent_select)
	    {
		in_extent_select = TRUE;
		SetCapture(hWnd);
		tty_mouse_select(MOUSE_SELECT_EXTENT, LOWORD(lParam), HIWORD(lParam));
	    }
	    return 0;
	}
	break;

    case WM_LBUTTONDBLCLK:
	if (allow_select)
	    tty_mouse_select(MOUSE_SELECT_WORD, LOWORD(lParam), HIWORD(lParam));
	return 0;

    case WM_RBUTTONDBLCLK:
	if (!has_3_button_mouse)
	    do_paste(hWnd);
	return 0;


    case WM_MOUSEMOVE:
	if (allow_select)
	{
	    if (in_select)
	    {
		tty_mouse_select(MOUSE_SELECT_MOVE, LOWORD(lParam), HIWORD(lParam));
	    }
	    else if (in_extent_select)
	    {
		tty_mouse_select(MOUSE_SELECT_EXTENT, LOWORD(lParam), HIWORD(lParam));
	    }
	}
	return 0;

    case WM_LBUTTONUP:
	if (allow_select && in_select)
	{
	    in_select = FALSE;
	    tty_mouse_select(MOUSE_SELECT_END, LOWORD(lParam), HIWORD(lParam));
	    ReleaseCapture();
	}
	return 0;

    case WM_RBUTTONUP:
	if (allow_select && has_3_button_mouse)
	{
	    if (in_extent_select)
	    {
		in_extent_select = FALSE;
		tty_mouse_select(MOUSE_SELECT_END, LOWORD(lParam), HIWORD(lParam));
		ReleaseCapture();
	    }
	    return 0;
	}
	break;

    case WM_MBUTTONUP:
	do_paste(hWnd);
	return 0;


    }
    return DefWindowProc(hWnd,msg,wParam,lParam);
}

static DWORD process_output(void)
{
    unsigned char	buff[4096];
    int			dread;

    while ((dread = recv(telnet_socket, buff, sizeof(buff) - 1, 0)) > 0)
    {
	term_need_refresh = FALSE;
	dread = tty_telnet_controls(buff, dread);
	emu_output_str(buff, dread);
	if (dread < 10)
	    tty_refresh(hwndMain, NULL);
	else
	    term_need_refresh = TRUE;
    }

    if (telnet_socket)
    {
	closesocket(telnet_socket);
	telnet_socket = 0;
    }

    allow_close = TRUE;
    PostMessage(hwndMain,WM_CLOSE,0,0);

    return 1;
}

static BOOL wsa_started = FALSE;

static BOOL host_connect(void)
{
    struct sockaddr_in	sin;
    struct hostent	*remote_host;
    DWORD		ThreadID;
    char		title[255];


    if (WSAStartup(MAKEWORD(1, 0), &WsaData) != 0)
    {
        MessageBox(hwndMain, "Could not find high enough version of WinSock", "Error", MB_OK | MB_ICONSTOP | MB_SETFOREGROUND);
        return FALSE;
    }

    wsa_started = TRUE;

    remote_host = gethostbyname(telnet_host);
    if (!remote_host)
    {
	MessageBox(hwndMain, "Unable to locate host", "Error",  MB_OK | MB_ICONSTOP | MB_SETFOREGROUND);
	WSACleanup();
	return FALSE;
    }

    telnet_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (telnet_socket == INVALID_SOCKET)
    {
	MessageBox(hwndMain, "Could not create socket", "Error",  MB_OK | MB_ICONSTOP | MB_SETFOREGROUND);
	WSACleanup();
	return FALSE;
    }

    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = htonl(INADDR_ANY);
    sin.sin_port = 0;
    bind(telnet_socket, (struct sockaddr *)&sin, sizeof(sin));

    sin.sin_family = AF_INET;
    sin.sin_port = htons(telnet_port);	/* The telnet port number */
    sin.sin_addr.S_un.S_un_b.s_b1= remote_host->h_addr[0];
    sin.sin_addr.S_un.S_un_b.s_b2= remote_host->h_addr[1];
    sin.sin_addr.S_un.S_un_b.s_b3= remote_host->h_addr[2];
    sin.sin_addr.S_un.S_un_b.s_b4= remote_host->h_addr[3];

    tty_output_str("Connecting to %s...", telnet_host);

    if (connect(telnet_socket, (struct sockaddr *)&sin, sizeof(sin)))
    {
	MessageBox(hwndMain, "Could not connect to host", "Error",  MB_OK | MB_ICONSTOP | MB_SETFOREGROUND);
	WSACleanup();
	return FALSE;
    }

    sprintf(title, "%s - %s", WINDOW_NAME, telnet_host);

    SetWindowText(hwndMain, title);

    if (CreateThread(
	     (LPSECURITY_ATTRIBUTES)NULL,           // No security attributes.
	     (DWORD)0,                              // Use same stack size.
	     (LPTHREAD_START_ROUTINE)process_output, // Thread procedure.
	     (LPVOID)NULL,                   	    // Parameter to pass.
	     (DWORD)0,                              // Run immediately.
	     (LPDWORD)&ThreadID)==NULL)
    {
	MessageBox(hwndMain, "Could not create thread", "Error",  MB_OK | MB_ICONSTOP | MB_SETFOREGROUND);
	WSACleanup();
	return FALSE;
    }


    return TRUE;

}

static void	get_colour(char *s, COLORREF *colour)
{
    int			r, g, b;

    if (sscanf(s, "%2x%2x%2x", &r, &g, &b) == 3)
    {
	*colour = RGB(r, g, b);
    }
}

static void	show_usage(char *message)
{
    tty_startup();
    tty_cursor_off();
    tty_output_str("Usage: vt [options] host [port-number]\r\n\r\n");
    tty_output_str("options:\r\n");
    tty_output_str("  -f{Font}[:size]  Specify font and size (in pixels)\r\n");
    tty_output_str("  -b{rrggbb}       Set the background colour\r\n");
    tty_output_str("  -c{rrggbb}       Set the text colour\r\n");
    tty_output_str("  -e{rrggbb}       Set the bold text colour\r\n");
    tty_output_str("  -k{rrggbb}       Set the cursor colour\r\n");
    tty_output_str("  -h{n}            Retain n lines of history\r\n");
    tty_output_str("  -v               Use variable size window\r\n");
    tty_output_str("  -s               Disable text selection\r\n");
    tty_output_str("  -n               Do not paste newline character from clipboard\r\n");
    tty_output_str("  -d               Disable the window close operation\r\n");
    tty_output_str("  -t{term}         Set the terminal type. default='%s'\r\n", terminal_type);
    tty_output_str("  -g{80x24}        Set the terminal size. default=80x24\r\n");
    tty_output_str("\r\n");
    tty_output_str("Examples:\r\n");
    tty_output_str("  vt -f'Courier New:20' -b00007f -c00ff00 -h100 host\r\n");

}

/**
 * -v			Use variable size window
 * -f'Courier New:20'	Use courier new font with size 20 pixels.
 * -b000080		Use blue background
 * -cff0000		Use red for normal text
 * -e00ff00		Use green for bold text
 * -k00ff00		Use green for cursor background
 */
static void	parse_cmdline(char *cmdline)
{
    char	*s;
    int		i;
    int		foreground_changed = FALSE;
    int		background_changed = FALSE;
    int		cursor_changed = FALSE;

    while (cmdline != NULL)
    {
	int	in_single_quote = FALSE;
	int	in_double_quote = FALSE;
	char	*arg = cmdline;

	while (*cmdline == ' ')
	    cmdline++;

	for (s = cmdline; *s != '\0'; s++)
	{
	    if (*s == ' ')
	    {
		if (in_single_quote || in_double_quote)
		    continue;
		*s = '\0';
		s++;
		break;
	    }
	    else if (*s == '\'')
	    {
		if (in_double_quote)
		    continue;
		in_single_quote = !in_single_quote;
		memmove(s, s + 1, strlen(s));
		s--;
		continue;
	    }
	    else if (*s == '\"')
	    {
		if (in_single_quote)
		    continue;
		in_double_quote = !in_double_quote;
		memmove(s, s + 1, strlen(s));
		s--;
		continue;
	    }
	    else if (*s == '\\')
	    {
		memmove(s, s + 1, strlen(s));
		s--;
		continue;
	    }
	}

	cmdline = s;

	while (*s == ' ')
	    cmdline++;

	if (cmdline[0] == '\0')
	    cmdline = NULL;

	if (arg[0] == '\0')
	    continue;

	if (arg[0] == '-' || arg[0] == '/')
	{
	    if (arg[1] == '\0')
		continue;

	    switch (arg[1])
	    {
	    case 'f':
		if (arg[2])
		{
		    char	*name = arg + 2;
		    char	*size = strchr(name, ':');
		    int		w, h;

		    if (size)
		    {
			if (sscanf(size + 1, "%dx%d", &w, &h) == 2 && w > 0 && h > 0)
			{
			    font_width = w;
			    font_height = h;
			}
			else if (sscanf(size + 1, "%d", &h) == 1 && h > 0)
			{
			    font_height = h;
			}
			*size = '\0';
		    }

		    if (*name)
			tty_fontname = name;
		}
		break;

	    case 'v':
		mode_fix_window_size = FALSE;
		break;

	    case 'n':
		paste_newline = FALSE;
		break;

	    case 'd':
		allow_close = FALSE;
		break;

	    case 's':
		allow_select = FALSE;
		break;

	    case 'h':
		max_screen_history = atoi(arg + 2);
		if (max_screen_history < 0)
		    max_screen_history = 0;
		if (max_screen_history > 5000)
		    max_screen_history = 5000;
		break;

	    case 't':
		if (arg[2])
		    terminal_type = strdup(arg + 2);
		break;

	    case 'g':
		if (arg[2])
		{
		    int		cols, rows;

		    if (sscanf(arg + 2, "%dx%d", &cols, &rows) == 2 &&
		    	cols > 10 && rows > 5)
		    {
			tty_set_size(cols, rows);
		    }
		}
		break;

	    case 'b':
		get_colour(arg + 2, &colour_background);
		background_changed = TRUE;
		break;

	    case 'c':
		get_colour(arg + 2, &colour_text);
		foreground_changed = TRUE;
		break;

	    case 'e':
		get_colour(arg + 2, &colour_bold);
		foreground_changed = TRUE;
		break;

	    case 'k':
		get_colour(arg + 2, &colour_cursor);
		cursor_changed = TRUE;
		break;
	    }

	    continue;
	}

	if (!telnet_host)
	    telnet_host = arg;
	else if ((i = atoi(arg)) > 0)
	    telnet_port = i;
    }


    /*
    ** If either the foreground or background has changed, then
    ** changed the cursor colour too.
    */
    if ((background_changed || foreground_changed) && !cursor_changed)
    {
	colour_cursor = colour_text;
	colour_cursor_text = colour_background;
    }

}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, INT nCmdShow)
{
	MSG 	msg;

	hInst = hInstance;

	if (!InitApplication())
	    return 0;

	emu_init();

	if ((hwndMain = CreateVtTermWndClassWnd()) == (HWND)0)
	    return 0;

	parse_cmdline(lpCmdLine);


	tty_startup();
	tty_init(hwndMain);
	ShowWindow(hwndMain,SW_SHOW);

	if (telnet_host == NULL)
	{
	    show_usage("No host specified");
	}
	else
	{
	    if (!host_connect())
		return 0;
	}

	SetTimer(hwndMain, IDTIMER, 100, NULL);

	while (GetMessage(&msg,NULL,0,0))
	{
	    if (msg.message == WM_KEYDOWN || msg.message == WM_KEYUP)
	    {
		if (!key_used(msg.wParam))
		    TranslateMessage(&msg);
	    }
	    DispatchMessage(&msg);
	}

	if (telnet_socket)
	    closesocket(telnet_socket);

	if (wsa_started)
	    WSACleanup();

	return msg.wParam;
}
